"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""
"""Driver for the Power Standards Lab pQube meter.  

The driver scrapes the html spit out by the meter, making it dependent
on a particular interface.

Config options:
Address : url of the meter
Rate [default 10] : polling period

"""


import sys
import logging
import time
import calendar
import threading
import urllib2
import httplib
import urlparse
import re
import signal

from smap.driver import SmapDriver
from smap.util import periodicSequentialCall

urllib2.install_opener(urllib2.build_opener())

PQUBE_POINTS = [
    ('L1-N', r'^(\d+\.\d+)', 'A', 'phase-neutral_voltage', 'V'), #ultiplier=1, divisor=1,
    ('L2-N', r'^(\d+\.\d+)', 'B', 'phase-neutral_voltage', 'V'), #ultiplier=1, divisor=1,
    ('L3-N', r'^(\d+\.\d+)', 'C', 'phase-neutral_voltage', 'V'), #ultiplier=1, divisor=1,

    ('L1 Amp', r'^(\d+\.\d+)', 'A', 'current', 'A'), #ultiplier=1, divisor=1,
    ('L2 Amp', r'^(\d+\.\d+)', 'B', 'current', 'A'), #ultiplier=1, divisor=1,
    ('L3 Amp', r'^(\d+\.\d+)', 'C', 'current', 'A'), #ultiplier=1, divisor=1,

#     ('L1-N Voltage Fundamental', r'^(\d+\.\d+)', 'A', 'a-n fundamental voltage'),
#     ('L1-N Voltage Fundamental', r'^(\d+\.\d+)', 'A', 'a-n fundamental phase'),
#     ('L2-N Voltage Fundamental', r'^(\d+\.\d+)', 'B', 'a-n fundamental voltage'),
#     ('L2-N Voltage Fundamental', r'^(\d+\.\d+)', 'B', 'a-n fundamental phase'),
#     ('L3-N Voltage Fundamental', r'^(\d+\.\d+)', 'C', 'a-n fundamental voltage'),
#     ('L3-N Voltage Fundamental', r'^(\d+\.\d+)', 'C', 'a-n fundamental phase'),

    ('Frequency', r'^(\d+\.\d+)', 'ABC', 'line_frequency', 'Hz'), #ultiplier=1, divisor=1,
    ('Voltage THD', r'^(\d+\.\d+)', 'ABC', 'voltage_thd', 'pct'), #ultiplier=1, divisor=1,
    ('Current TDD', r'^(\d+\.\d+)', 'ABC', 'current_tdd', 'pct'), #ultiplier=1, divisor=1,

    ('L1-L2', r'^(\d+\.\d+)', 'AB', 'volts', 'V'), #ultiplier=1, divisor=1,
    ('L2-L3', r'^(\d+\.\d+)', 'BC', 'volts', 'V'), #ultiplier=1, divisor=1,
    ('L3-L1', r'^(\d+\.\d+)', 'AC', 'volts', 'V'), #ultiplier=1, divisor=1,

    ('Power', r'^(\d+\.\d+)', 'ABC', 'true_power', 'kW'), #ultiplier=1, divisor=1,
    ('Apparent Power', r'^(\d+\.\d+)', 'ABC', 'apparent_power', 'kVA'), #ultiplier=1, divisor=1,
    ('Reactive Power', r'^(\d+\.\d+)', 'ABC', 'reactive_power', 'kVAR'), #ultiplier=1, divisor=1,
    ('True Power Factor', r'^(\d+\.\d+)', 'ABC', 'pf', 'PF'), #ultiplier=1, divisor=1,

    # meters
    ('Energy', r'^(\d+\.\d+)', 'ABC', 'true_energy', 'kWh'), #ultiplier=1000, divisor=1,
    ('Apparent Energy', r'^(\d+\.\d+)', 'ABC', 'apparent_energy', 'kVAh'), #ultiplier=1000, divisor=1,
    ]
class PQube(SmapDriver):
    def setup(self, opts):
        self.serverloc = opts['Address']
        self.rate = opts.get('Rate', 10)
        for (field, regexp, phase, channel, fmt) in PQUBE_POINTS:
            self.add_timeseries('/%s/%s' % (phase, channel), fmt, data_type="double")
            self.set_metadata('/%s' % phase, {
                'Extra/Phase' : phase})

        self.set_metadata('/', {
            'Instrument/Manufacturer' : 'Power Standards Laboratory',
            'Instrument/SamplingPeriod' : str(self.rate),
            'Extra/Driver' : 'smap.drivers.pqube.PQube',
            })

    def start(self):
        print "satrtin"
        periodicSequentialCall(self.update).start(self.rate)
        
    def update(self):
        logging.debug("Updating " + self.serverloc)
        try:
            print "loading", self.serverloc + '/Meters.htm'
            fp = urllib2.urlopen(self.serverloc + '/Meters.htm', timeout=15)
            html = fp.read()
        except IOError, e:
            logging.error("IOError while reading pqube: url: %s exception: %s" % (url, str(e)))
            return
        except httplib.HTTPException, e:
            logging.error("HTTP exception reading pqube: url: %s exception: %s" % (url, str(e)))
            return

        reading_time = int(time.time())
        # this pulls out a list of all the channel-reading pairs
        data = re.findall('<td.*? class="channel">(.*?)</td>.*?<td.*?>(.*?)</td>',
                          html.lower())
        data = [(re.sub('<.*>', '', k), v) for (k,v) in data]
        data_map = {}
        data_map.update(data)

        for (field, regexp, phase, channel, fmt) in PQUBE_POINTS:
            reading = data_map.get(field.lower())
            if not reading: 
                logging.warn(field + " not found in doc")
                continue
            match = re.search(regexp, reading)
            if not match: 
                logging.warn("reading conversion failed: " + reading)
                continue

            self.add('/%s/%s' % (phase, channel), reading_time, float(match.groups(0)[0]))




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import util

import os
import shutil
import unittest

def snp(s):
    return "%08i" % s

class DiskLog:
    """Class which keeps an on-disk log of records
    """
    def _write_meta(self):
        util.pickle_dump(os.path.join(self.dirname, 'META'), self.meta)

    def _read_meta(self):
        self.meta = util.pickle_load(os.path.join(self.dirname, 'META'))

    def _write_tail(self):
        util.pickle_dump(os.path.join(self.dirname, snp(self.meta['tail'] - 1)), self._tail)

    def _read_seqno(self, seq):
        try:
            return util.pickle_load(os.path.join(self.dirname, snp(seq)))
        except IOError:
            return None

    def __init__(self, dirname):
        self.dirname = dirname

        if not os.path.isdir(dirname):            
            # create a new log
            os.makedirs(dirname)
            self.meta = {
                'head' : 0,
                'tail' : 0
                }
            self._write_meta()
            self._head = self._tail = None
        else:
            # read the head and tail off disk for an existing log
            self._read_meta()
            if self.meta['tail'] > 0:
                self._tail = self._read_seqno(self.meta['tail'] - 1)
            else:
                self._tail = None

            self._head = self._read_seqno(self.meta['head'])

        self.dirty = False
            
    def __len__(self):
        return self.meta['tail'] - self.meta['head']

    def tail(self):
        """Return the tail of the log"""
        return self._tail

    def head(self):
        return self._head

    def update_tail(self, obj):
        # change the value of the tail
        self._tail = obj

        # if we're length 1, that's also the head
        if self.meta['tail'] == self.meta['head'] + 1:
            self._head = obj

        # need to sync
        self.dirty = True
    
    def append(self, obj):
        # flush the current tail to disk
        self.sync()

        # add the new tail (and head if we were empty)
        self._tail = obj
        if self.meta['tail'] == self.meta['head']:
            self._head = obj
        self.meta['tail'] += 1

        # need a flush
        self.dirty = True

    def sync(self):
        if self.dirty:
            self._write_tail()
            self._write_meta()
            self.dirty = False

    close = sync

    def pop(self):
        """Truncate sequence numbers less than `seqno`
        """
        self.sync()

        try:
            os.remove(os.path.join(self.dirname, snp(self.meta['head'])))
        except OSError:
            pass

        if self.meta['tail'] > self.meta['head']:
            self.meta['head'] += 1

        if self.meta['tail'] == self.meta['head']:
            # q is now empty
            self._head = self._tail = None
        elif self.meta['tail'] == self.meta['head'] + 1:
            # q now has length 1.  grab the head since it might be dirty
            self._head = self._tail
        else:
            # read the new head off disk
            self._head = self._read_seqno(self.meta['head'])

    def idxtoseq(self, idx):
        pass


class TestDiskLog(unittest.TestCase):
    def test_1(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO")
            self.assertEqual(dl.head(), "FOO")
            self.assertEqual(dl.tail(), "FOO")
        finally:
            shutil.rmtree("testdir")

    def test_save_one(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO")
            self.assertEqual(dl.head(), "FOO")
            self.assertEqual(dl.tail(), "FOO")

            dl2 = DiskLog("testdir")
            self.assertEqual(dl.head(), "FOO")
            self.assertEqual(dl.tail(), "FOO")
        finally:
            shutil.rmtree("testdir")
        
    def test_append(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO 0")
            for i in xrange(1, 10):
                self.assertEqual(dl.tail(), "FOO " + str(i-1))
                self.assertEqual(dl.head(), "FOO 0")
                dl.append("FOO " + str(i))
        finally:
            shutil.rmtree("testdir")

    def test_trunc(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO 0")
            for i in xrange(1, 10):
                self.assertEqual(dl.tail(), "FOO " + str(i-1))
                self.assertEqual(dl.head(), "FOO 0")
                dl.append("FOO " + str(i))

            for i in xrange(1, 10):
                dl.pop()
                self.assertEqual(dl.head(), "FOO " + str(i))
        finally:
            shutil.rmtree("testdir")

    def test_trunc_split(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO 0")
            for i in xrange(1, 10):
                self.assertEqual(dl.tail(), "FOO " + str(i-1))
                self.assertEqual(dl.head(), "FOO 0")
                dl.append("FOO " + str(i))
            dl.sync()
                
            dl2 = DiskLog("testdir")
            for i in xrange(1, 10):
                dl2.pop()
                self.assertEqual(dl2.head(), "FOO " + str(i))
        finally:
            shutil.rmtree("testdir")
            
    def test_race(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO 0")
            for i in xrange(1, 10):
                self.assertEqual(dl.head(), "FOO " + str(i-1))
                self.assertEqual(dl.tail(), "FOO " + str(i-1))
                dl.append("FOO " + str(i))
                dl.pop()
            self.assertEqual(dl.head(), "FOO " + str(9))
            self.assertEqual(dl.tail(), "FOO " + str(9))
            dl.pop()
            self.assertEqual(dl.head(), None)
            self.assertEqual(dl.tail(), None)
        finally:
            shutil.rmtree("testdir")
        
    def test_edithead(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO")
            self.assertEqual(dl.head(), "FOO")
            self.assertEqual(dl.tail(), "FOO")
            dl.update_tail("FOO BAR")
            self.assertEqual(dl.head(), "FOO BAR")
            self.assertEqual(dl.tail(), "FOO BAR")
        finally:
            shutil.rmtree("testdir")
        
    def test_edithead_2(self):
        try:
            dl = DiskLog("testdir")
            dl.append("FOO")
            self.assertEqual(dl.head(), "FOO")
            self.assertEqual(dl.tail(), "FOO")
            dl.append("FOO2")
            self.assertEqual(dl.tail(), "FOO2")
            dl.update_tail("FOO BAR")
            self.assertEqual(dl.tail(), "FOO BAR")

            dl.pop()
            self.assertEqual(dl.head(), "FOO BAR")
            self.assertEqual(dl.tail(), "FOO BAR")

        finally:
            shutil.rmtree("testdir")

    def test_pop(self):
        try:
            dl = DiskLog("testdir")
            dl.append("1")
            dl.append("2")
            self.assertEqual(dl.head(), "1")
            self.assertEqual(dl.tail(), "2")
            dl.pop()
            self.assertEqual(dl.head(), "2")
            self.assertEqual(dl.tail(), "2")
            dl.pop()
            self.assertEqual(dl.head(), None)
            self.assertEqual(dl.tail(), None)
        finally:
            shutil.rmtree("testdir")

    def test_big(self):
        try:
            dl = DiskLog("testdir")
            for i in xrange(0, 10):
                dl.append([i for i in xrange(0, 100000)])
        finally:
            shutil.rmtree("testdir")

        
if __name__ == '__main__':
    unittest.main()



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

from twisted.web import resource, server
from twisted.python import log

from smap.core import SmapException
from smap.server import setResponseCode
import smap.util as util
import queryparse as qp

def receive_object(request, key, public):
    """Check if a given client should receive an object with
    associated key and public setting.
    """
    if 'key' in request.args and key in request.args['key']:
        return True
    elif not 'private' in request.args and public:
        return True
    else:
        return False

class ReResource(resource.Resource):
    """Provide a "republish resource" -- where you can long-poll to
    listen for new data.
    """
    def __init__(self, db):
        resource.Resource.__init__(self)
        self.listeners = {}
        self.db = db

    def connectionLost(self, request, reason):
        if request in self.listeners:
            log.msg("removing republish client")
            del self.listeners[request]

    def render_GET(self, request):
        self.add_client(None, request)
        return server.NOT_DONE_YET

    def render_POST(self, request):
        parser = qp.QueryParser(request)
        query = "select distinct uuid where (%s)" % request.content.read() 
        try:
            d = parser.runquery(self.db, query)
        except SmapException, e:
            setResponseCode(request, e, 400)
            return "query: %s\nerror: %s\n" % (query, str(e))
        else:
            d.addCallback(self.add_client, request)
            d.addErrback(lambda _: request.finish())
            return server.NOT_DONE_YET

    def add_client(self, uuids, request):
        # make sure we remove the client when she disconnects
        request.notifyFinish().addErrback(lambda x: self.connectionLost(request, x))
        if uuids != None:
            self.listeners[request] = set(uuids)
            log.msg("adding republish client, %i topics" % len(uuids))
        else:
            self.listeners[request] = None
            log.msg("adding republish client, all topics" )
        return server.NOT_DONE_YET

    def republish(self, key, public, obj):
        data = util.json_encode(obj)
        for client, streams in self.listeners.iteritems():
            if receive_object(client, key, public):
                if streams == None:
                    # if they've subscribed to all streams, we can
                    # just forward them the object
                    client.write("\n\n")
                    client.write(data)
                else:
                    # only include topical data
                    custom = dict((k, obj[k]) for k in obj.iterkeys() \
                                      if not 'uuid' in obj[k] or  \
                                      obj[k]['uuid'] in streams)
                    # don't bother filtering metadata at the moment
                    # since it's expensive to construct and mostly
                    # won't happen.
                    if sum((1 for v in custom.itervalues() if 'uuid' in v)):
                        client.write(util.json_encode(custom))
                        client.write("\n\n")
                    



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import uuid
import datetime
import time
import sys
from twisted.internet import reactor, threads, defer
from zope.interface import implements

from interface import *
import core
import util
import loader
import server
from smap.contrib import dtutil

class SmapDriver:
    # this is actually a shim layer which presents a ISmapInstance to
    # drivers
    implements(ISmapInstance)
    load_chunk_size = datetime.timedelta(days=1)

    @classmethod
    def get_driver(cls, inst, name, attach_point, namespace):
        """Create a managed driver which will manage a driver whose
        implementation is named by "name"
        """
        cmps = name.split('.')
        assert len(cmps) > 1
        (mod_name, class_name) = ('.'.join(cmps[:-1]), cmps[-1])

        if mod_name in sys.modules:
            mod = sys.modules[mod_name]
        else:
            mod = __import__(mod_name, globals(), locals(), [class_name]) 

        klass = getattr(mod, class_name)
        driver = klass(inst, attach_point, namespace)
        inst.add_driver(attach_point, driver)
        return driver

    flush = lambda self: self.__inst.reports.flush()
    _flush = lambda self: self.__inst.reports._flush()

    def __init__(self, smap_instance, attach_point, namespace):
        self.__inst = smap_instance
        self.__attach_point = attach_point
        self.namespace = namespace

    def __join_id(self, id):
        if util.is_string(id) and id.startswith('/'):
            return util.norm_path(self.__attach_point + '/' + id)
        else:
            return id

    # override
    def setup(self, opts={}):
        pass

    # override
    def start(self):
        pass

    # ISmapInstance implementation

    # drivers get a pass-through version of the SmapInstance which
    # knows where they are attached in the tree.  Drivers may, for the
    # most part, assume that they have the instance to themselves.
    # However, the main exception to this is that their names are
    # still in a global namespace.  Therefore they should either
    # generate their own uuids for their streams based on the uuid in
    # their root collection (inst.get_collection('/')['uuid'])
    def lookup(self, id, **kwargs):
        return self.__inst.lookup(self.__join_id(id), **kwargs)
    def get_timeseries(self, id):
        return self.__inst.get_timeseries(self.__join_id(id))
    def get_collection(self, id):
        return self.__inst.get_collection(self.__join_id(id))
    def add_timeseries(self, path, *args, **kwargs):
        kwargs['namespace'] = self.namespace
        if len(args) <= 1:
            key = path
        elif len(args) == 2:
            key = args[0]
            args = args[1:]
        return self.__inst.add_timeseries(self.__join_id(path), key, *args, **kwargs)
    def add_actuator(self, path, unit, klass, **kwargs):
        return self.add_timeseries(path, unit, klass=klass, **kwargs)
    def add_collection(self, path, *args):
        self.__inst.add_collection(self.__join_id(path), *args)
    def set_metadata(self, id, *metadata):
        return self.__inst.set_metadata(self.__join_id(id), *metadata)
    def add(self, id, *args):
        return self.__inst.add(self.__join_id(id), *args)
    def _add(self, id, *args):
        return self.__inst._add(self.__join_id(id), *args)
    def uuid(self, key):
        return self.__inst.uuid(key, namespace=self.namespace)

    def load(self, startdt, enddt):
        """Default load method tries to call update with one-day chunks"""
        self.start_dt = startdt
        self.end_dt = enddt
        return self._load_time_chunk(self)

    def _load_time_chunk(self, *args):
        if self.start_dt >= self.end_dt:
            return None

        # pick a new window
        start = self.start_dt
        end = self.start_dt + self.load_chunk_size
        if end > self.end_dt: end = self.end_dt

        print "loading", self.start_dt, '-', self.end_dt
        self.start_dt = self.start_dt + self.load_chunk_size
        print start, end
        d = defer.maybeDeferred(self.update, start, end)
        d.addCallback(self.update)
        d.addCallback(lambda _: self._flush())
        d.addCallback(self._load_time_chunk)

        def err(e):
            print e
        d.addErrback(err)

        return d


class BaseDriver(SmapDriver):
    def setup(self, opts={}):
        self.t = self.add_timeseries('/sensor0', 'mytimeseries', 'SDH')
        self.set_metadata('/sensor0', { 
            'Instrument/ModelName' : 'ExampleInstrument',
            'Extra/ModbusAddr' : opts.get('ModbusAddr', '')
            })
        self.counter = int(opts.get('StartVal', 0))

    def start(self):
        self.counter = 0
        util.periodicSequentialCall(self.read).start(1)

    def read(self):
        # print "Add", self.counter
        self.t.add(self.counter)
        self.counter += 1
        # print self.counter


if __name__ == '__main__':
    # inst = loader.load('default.ini', autoflush=False)
    # d = SmapDriverManager.get_driver('driver.BaseDriver', '/newsensor')
    # d.setup(inst)
    import uuid
    inst = core.SmapInstance(uuid.uuid1())
    bd = SmapDriver.get_driver(inst, 'driver.BaseDriver', '/newsensor', None)
    bd.setup()

    server.run(inst)



# coding: utf-8
# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
# See LICENSE for details.

"""
See twisted/application/internet.py for details related to this module.

This module is part of the twistedpatch that attempts to provide IPv6 support
for Twisted without modifying the core Twisted files. The IPv6 patch is
applied at runtime and requires only minor modifications to your sources,
specifically to include support for this patch.

Be aware that this patch is neither supported nor endorsed by Twisted Matrix
Laboratories. Using this patch may break things. You have been warned.
"""

from twisted.application.internet import _AbstractClient
from twisted.application.internet import _AbstractServer as _TwistedAbstractServer
from twisted.application.internet import _VolatileDataService

class _AbstractServer (_TwistedAbstractServer):
    def _getPort (self):
        if self.reactor is None:
            from tx.ipv6.internet import reactor
        else:
            reactor = self.reactor
        return getattr(reactor, 'listen%s' % (self.method,))(*self.args,
                                                             **self.kwargs)
        
_doc={
'Client':
"""Connect to %(tran)s

Call reactor.connect%(method)s when the service starts, with the
arguments given to the constructor.
""",
'Server':
"""Serve %(tran)s clients

Call reactor.listen%(method)s when the service starts, with the
arguments given to the constructor. When the service stops,
stop listening. See twisted.internet.interfaces for documentation
on arguments to the reactor method.
""",
}
        
import new
for tran in 'TCP TCP6 UNIX SSL UDP UNIXDatagram Multicast'.split():
    for side in 'Server Client'.split():
        if tran == "Multicast" and side == "Client":
            continue
        base = globals()['_Abstract'+side]
        method = {'Generic': 'With'}.get(tran, tran)
        doc = _doc[side]%vars()
        klass = new.classobj(tran+side, (base,),
                             {'method': method, '__doc__': doc})
        globals()[tran+side] = klass
        
__all__ = (['TimerService', 'CooperatorService', 'MulticastServer'] +
           [tran+side
            for tran in 'Generic TCP UNIX SSL UDP UNIXDatagram'.split()
            for side in 'Server Client'.split()])


"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import os
import uuid
from zope.interface import implements
from twisted.web import resource
from twisted.internet import reactor, defer
import exceptions
import sys
import operator

import schema
import util
import reporting
import smapconf
from interface import *

class SmapException(Exception):
    """Generic error"""
    def __init__(self, message, http_code=None):
        Exception.__init__(self, message)
        self.http_code = http_code

class SmapSchemaException(SmapException):
    """Exception generated if a json object doesn't validate as the
appropriate kind of schema"""

class Timeseries(dict):
    """Represent a single Timeseries.  A Timeseries is a single stream of
    scalars, with associated units.

    The sMAP profile requires each time series to be associated with a
    data type (long, double, or string), unit of measure, and
    timezone.
    """
    implements(ITimeseries)

    FIELDS = ["Readings", "Description", "Metadata", 
              "Properties", "uuid"]

    # default values for the initializer  
    # these are used both here and when the loader module creates a
    # sMAP instance from a config file
    DEFAULTS = {
        'BufferSize' : 1,
        'Properties/Timezone' : 'America/Los_Angeles',
        'Properties/ReadingType' : 'long'
        }

    def __init__(self,
                 new_uuid,
                 unit, 
                 data_type=DEFAULTS['Properties/ReadingType'],
                 timezone=DEFAULTS['Properties/Timezone'],
                 description=None,
                 buffersz=DEFAULTS['BufferSize'],
                 milliseconds=False):
        """
:param new_uuid: a :py:class:`uuid.UUID`
:param string unit: the engineering units of this timeseries
:param string data_type: the data type of the data.  options are ``long``, ``double``, and ``string``.
:param string timezone: a tzinfo-style timezone.
:param string description: the value of sMAP Description field.
:param int buffersz: how many readings to present when the timeseries is retrieved with a ``GET``.
:param bool milliseconds: if True, then the stream publishes time in
 units of Unix milliseconds.  Otherwise, normal unix timestamps are
 assumed
"""
        if isinstance(new_uuid, dict):
            if not schema.validate('Timeseries', new_uuid):
                raise SmapSchemaException("Initializing timeseries failed -- invalid object")
            dict.__init__(self, new_uuid)
            reading_init = new_uuid['Readings']
        else:
            self.__setitem__("uuid", new_uuid)
            self.__setitem__("Properties", {
                    'UnitofMeasure' : unit,
                    'ReadingType' : data_type,
                    'Timezone' : timezone})
            if description:
                self.__setitem__("Description", description)
            reading_init = []
        self.dirty = True
        self.milliseconds = milliseconds
        self.__setitem__("Readings", util.FixedSizeList(buffersz, init=reading_init))

    def _check_type(self, value):
        type_ = self.__getitem__('Properties')['ReadingType']
        if type_ == 'long' and util.is_integer(value):
            return True
        elif type_ == 'double' and \
                isinstance(value, float):
            return True
        else:
            return False

    def _add(self, *args):
        """Add a new reading to this timeseries.  This version must
only be called from the :py:mod:`twisted` main loop; *i.e.* from a
callback added with ``reactor.callFromThread()``

Can be called with 1, 2, or 3 arguments.  The forms are

* ``_add(value)``
* ``_add(time, value)``
* ``_add(time, value, seqno)``

:raises SmapException: if the value's type does not match the stream
 type, or was called with an invalid number of arguments.
        """
        seqno = None
        if len(args) == 1:
            time = util.now()
            value = args[0]
        elif len(args) == 2:
            time, value = args
        elif len(args) == 3:
            time, value, seqno = args
        else:
            raise SmapException("Invalid add arguments: must be (value), "
                                "(time, value), or (tiem, value, seqno)")
        time = int(time)
        if not self.milliseconds:
            time *= 1000

        if not self._check_type(value):
            raise SmapException("Attempted to add " + str(value) + 
                                " to Timeseries, but " +
                                "the timeseries type is " + 
                                self.__getitem__('Properties')['ReadingType'])
        
        if seqno: reading = time, value, seqno
        else: reading = time, value
        self["Readings"].append(reading)
        if not hasattr(self, 'inst'): return

        # if a timeseries is dirty, we need to republish all of its
        # metadata before we publish it so stream is right. some of
        # this may have already been published, in which case it won't
        # actually do anything.
        if self.dirty:
            split_path = util.split_path(getattr(self, 'path'))
            for i in xrange(0, len(split_path)):
                path_seg = util.join_path(split_path[:i])
                self.inst.reports.publish(path_seg, 
                                          self.inst.get_collection(path_seg))
            rpt = dict(self)
            rpt['Readings'] = [reading]
            self.inst.reports.publish(getattr(self, 'path'), rpt)
            self.dirty = False
        else:
            # publish a stripped-down Timeseries object
            self.inst.reports.publish(getattr(self, 'path'),
                                      {'uuid' : self['uuid'],
                                       'Readings' : [reading]})

    def add(self, *args):
        """A version of :py:meth:`~Timeseries._add` which can be called from any thread.
        """
        # SDH : thread-safe
        # this way the real add is always done in the main loop,
        # even if it was called by another threadpool or something.
        reactor.callFromThread(lambda: self._add(*args))

    def __setitem__(self, attr, value):
        if attr in self.FIELDS:
            dict.__setitem__(self, attr, value)
            if attr != 'uuid':
                if not schema.validate("Timeseries", self):
                    raise SmapSchemaException("Invalid schema in " 
                                              "Timeseries for " + 
                                              attr)
            # time series start dirty so when we publish them the
            # first time we send all their metadata.
            self.dirty = True
        else:
            raise KeyError(attr + " can not be set on a Timeseries!")

    def set_metadata(self, metadata):
        metadata = util.build_recursive(metadata)
        if 'Metadata' in metadata:
            metadata = metadata['Metadata']
        self['Metadata'] = util.dict_merge(self.get('Metadata', {}),
                                           metadata)

    
    def render(self, request):
        return self

class Collection(dict):
    """Represent a collection of sMAP resources"""
    implements(ICollection)
    def __init__(self, path, inst=None, description=None, *args):
        """
        :param string path: the path where the collection will be added
        :param SmapInstance inst: the containing :py:class:`SmapInstance` object
        :param string description: the contents of the sMAP description field
        :raise SmapSchemaException: if the resulting object does not validate
        """
        self.inst = inst
        setattr(self, 'path', util.norm_path(path))
        if len(args) == 1 and isinstance(args[0], dict):
            dict.__init__(self, args[0])
        else:
            self.__setitem__("Contents", [])

        if not schema.validate("Collection", self):
            raise SmapSchemaException("Error instantiating Collection: " 
                                      "invalid parameter")

    def add_child(self, name):
        """Add a child name to the collection
        """
        self["Contents"].append(name)

    def dirty_children(self):
        """Recursively mark all timeseries contained in this collection as dirty
        """
        def explore(item, path):
            if not 'Contents' in item:
                item.dirty = True
            else:
                for ps in item['Contents']:
                    newpath = path + [ps]
                    explore(self.inst.lookup(util.join_path(newpath)), 
                            newpath)
        if hasattr(self, 'path'):
            explore(self, util.split_path(getattr(self, 'path')))

    def __setitem__(self, attr, value):
        if not attr in ['Contents', 'Metadata', 'Proxy']:
            raise SmapException("Key " + attr + " cannot be set on a Collection!")
        elif not attr in self or value != dict.__getitem__(self, attr):
            self.dirty_children()
            dict.__setitem__(self, attr, value)
            self.dirty_children()

    def update(self, val):
        if 'Contents' in val:
            del val['Contents']
        dict.update(self, val)

    def set_metadata(self, metadata):
        metadata = util.build_recursive(metadata)
        if 'Metadata' in metadata:
            metadata = metadata['Metadata']
        self['Metadata'] = util.dict_merge(self.get('Metadata', {}),
                                           metadata)
        self.dirty_children()

    def render(self, request):
        return self


class SmapInstance:
    """A sMAP instance is a tree of :py:class:`Collections` and
:py:class:`Timeseries`.  A :py:class:`SmapInstance` allows lookups
based on either path or UUID, and also contains a reference to the
sMAP reporting functionality."""
    implements(ISmapInstance)

    def __init__(self, root_uuid, **kwargs):
        if not isinstance(root_uuid, uuid.UUID):
            root_uuid = uuid.UUID(root_uuid)

        self.OBJS_PATH = {}
        self.OBJS_UUID = {}
        self.drivers = {}

        # if we're not given an explicit report file, put it in the
        # datadir or else the cwd
        if not 'reportfile' in kwargs:
            if 'datadir' in smapconf.SERVER:
                rf = os.path.join(smapconf.SERVER['datadir'], str(root_uuid))
            else:
                rf = str(root_uuid)

            self.reports = reporting.Reporting(self, reportfile=rf, **kwargs)
        else:
            self.reports = reporting.Reporting(self, **kwargs)
        self.flush = self.reports.flush
        self._flush = self.reports._flush
        self.loading = False
        self.add_collection("/")
        self.root_uuid = root_uuid

    # keep a list of sensor drivers so we can find them easily
    def add_driver(self, path, driver):
        self.drivers[path] = driver

    def start(self):
        """Causes the reporting subsystem and any drivers to be started
        """
        map(lambda x: x.start(), self.drivers.itervalues())

    def uuid(self, key, namespace=None):
        if not namespace:
            namespace = self.root_uuid
        if key and namespace:
            rv = uuid.uuid5(namespace, key)
            if rv in self.OBJS_UUID:
                raise SmapException("Duplicate UUID detected -- this probably "
                                    "means you specifed the same key twice in "
                                    "the same namespace.  The offending key is " + 
                                    str(key))
            return rv
        else:
            raise SmapException("Timeseries cannot generate uuid: must "
                            "specifiy either uuid or key and ROOT_UUID")

    def lookup_path(self, id):
        """Look up the path of Timeseries by uuid
        :param uuid.UUID id: the id of the timeseries in question
        """
        t = self.get_timeseries(id)
        if t and hasattr(t, 'path'):
            return getattr(t, 'path')
        return None
    
    def lookup(self, id, pred=None):
        """Retrieve an object in the resource hierarchy by path or uuid.  If
        *id* is a string not starting with ``/``, it will be passed to the
        :py:class:`uuid.UUID` constructor; otherwise it will be treated as a
        pathname.  *pred* is an optional predicate which can be used to test
        the result.
"""
        if util.is_string(id):
            path = util.split_path(id)
            if len(path) > 0 and path[-1][0] == "+":
                return self._lookup_r(util.join_path(path[:-1]), pred=pred)
            else:
                obj = self.OBJS_PATH.get(util.join_path(path), None)
        elif isinstance(id, uuid.UUID):
            return self.OBJS_UUID.get(id, None)
        else:
            obj = None
        
        if not pred or pred(obj):
            return obj
        else: return None

    def _lookup_r(self, id, pred=None):
        """Lookup recursively in the resource hierarchy, starting with the
        resource identifed by "id".  Returns a list of elements for which
        "pred" returns True"""
        rv = {}
        q = [id]
        root_path = getattr(self.lookup(id), 'path')

        while len(q) > 0:
            cur = self.lookup(q.pop(0))
            if ICollection.providedBy(cur):
                for child in cur['Contents']:
                    q.append(getattr(cur, 'path') + '/' + child)
            if cur and (not pred or pred(cur)):
                rvpath = util.norm_path(getattr(cur, 'path')[len(root_path):])
                rv[rvpath] = cur
        return rv

    @staticmethod
    def render_lookup(request, val):
        """Render a return value of lookup, by calling render()
        methods on all the results.  Returns either the object or a
        deferred that you must wait on for the result to be ready.
        """
        if hasattr(val, "render") and callable(val.render):
            # if its not a collection of objects, just render it
            return val.render(request)
        elif isinstance(val, dict):
            # otherwise we might need to render() multiple objects and
            # wait for all of them -- this can happen if we are
            # reading multiple actuators.
            rv = {}
            deferreds = []

            # start all the rendering
            for k, v in val.iteritems():
                rendered = v.render(request)
                if isinstance(rendered, defer.Deferred):
                    deferreds.append((k, rendered))
                else:
                    rv[k] = rendered

            # set up a deferred task which will fire when all results are ready
            if len(deferreds) == 0:
                return rv
            else:

                # and insert the ones which succeeded into the result
                # set.
                def insertResults(vals):
                    for ((path, d), (success, yld)) in zip(deferreds, vals):
                        if success:
                            rv[path] = yld
                    return rv

                d = defer.DeferredList(map(operator.itemgetter(1), deferreds), consumeErrors=True)
                d.addCallback(insertResults)
                return d

    def get_timeseries(self, path): 
        """Returns a :py:class:`Timeseries` if an object is found
        matching *path*, or None otherwise."""
        return self.lookup(path, pred=ITimeseries.providedBy)

    def get_collection(self, path):
        """Returns a :py:class:`Collection` if an object with an
        identifier matching *path* is found, None otherwise
        """
        return self.lookup(path, pred=ICollection.providedBy)

    def _add(self, path, *args, **kwargs):
        """Utility to call the version of :py:meth:`~smap.core.Timeseries._add`
        associated with *path*.  The same as ``inst.get_timeseries(path)._add(...)``
        """
        return self.get_timeseries(path)._add(*args, **kwargs)
        
    def add(self, path, *args, **kwargs):
        """Utility to call the version of :py:meth:`~smap.core.Timeseries.add`
        associated with *path*.  The same as ``inst.get_timeseries(path).add(...)``
        """
        try:
            return self.get_timeseries(path).add(*args, **kwargs)
        except exceptions.AttributeError, e:
            raise SmapException("add failed: no such path: %s" % path)

    def _add_parents(self, path):
        for i in xrange(0, len(path)):
            if not self.get_collection(util.join_path(path[:i])):
                self.add_collection(util.join_path(path[:i]))

    def add_timeseries(self, path, *args, **kwargs):
        """Add a timeseries to the smap server at the given path.  This will
        generate a UUID for the timeseries.

        direct form 
        :param path a Timeseries instance
        
        simple form 
        :param args[0] is a uuid instance, or a key to generate a uuid with by combining it with the root uuid.
        :param args[1] and kwargs are arguments passed to the Timeseries constructor.  Therefore you have to include at least the UnitofMeasure

        :param boolean replace: (kwarg) replace an existing timeseries at that path instead of throwing an exception
        :param boolean recurse: recursively create parent collections instead of thrwoing an exception.  Default is True.

        :raises: :py:class:`SmapException` if the parent isn't a collection or the path already exists.
        """ 
        replace = kwargs.pop('replace', False)
        recurse = kwargs.pop('recurse', True)
        klass = kwargs.pop('klass', Timeseries)

        if len(args) == 0 or \
                not ITimeseries.providedBy(args[0]) and not IActuator.providedBy(args[0]):
            if len(args) == 2:
                if not isinstance(args[0], uuid.UUID):
                    id = self.uuid(args[0], namespace=kwargs.get('namespace', None))
                else:
                    id = args[0]
                args = args[1:]
            elif len(args) == 1:
                id = self.uuid(util.norm_path(path), kwargs.get('namespace', None))
            else:
                id = self.uuid(util.norm_path(path))
#                 raise SmapException("SmapInstance.add_timeseries may only be called "
#                                     "with two or three arguments")

            kwargs.pop('namespace', None)
            timeseries = klass(id, *args, **kwargs)
            if id != args[0]:
                setattr(timeseries, "key", args[0])
        else:
            timeseries = args[0]

        path = util.split_path(path)
        if recurse: self._add_parents(path)
        parent = self.get_collection(util.join_path(path[:-1]))

        if not replace and util.join_path(path) in self.OBJS_PATH:
            raise SmapException("add_timeseries: path " + str(path) + " exists!")
        if not parent:
            raise SmapException("add_timeseries: parent is not a collection!")
        parent.add_child(path[-1])

        # place the new timeseries into the uuid and path tables
        self.OBJS_UUID[timeseries['uuid']] = timeseries
        self.OBJS_PATH[util.join_path(path)] = timeseries
        timeseries.inst = self
        setattr(timeseries, 'path', util.join_path(path))
        if not self.loading: self.reports.update_subscriptions()
        return timeseries

    def add_collection(self, path, *args): 
        """Add collection to the namespace.  For instance::

          inst.add_collection('/c1')

        :param string path: path under which to add the collection
        :param args[0]: :py:class:`Collection` class to add, if present
        :rtype: the :py:class:`Collection` which was added
        :raises: :py:class:`SmapException` if the parent is not a
         collection, or the path exists.
"""
        if len(args) > 0 and ICollection.providedBy(args[0]):
            collection = args[0]
        elif len(args) == 0:
            collection = Collection(path, self)
        else:
            raise SmapException("add_collection: wrong number of arguments")

        path = util.split_path(path)
        if len(path) > 0:
            parent = self.get_collection(util.join_path(path[:-1]))
            if not parent:
                raise SmapException("add_collection: parent is not collection!")
            parent.add_child(path[-1])
        if util.join_path(path) in self.OBJS_PATH:
            raise SmapException("add_timeseries: path " + str(path) + 
                                " exists!")

        self.OBJS_PATH[util.join_path(path)] = collection
        if not self.loading: self.reports.update_subscriptions()
        return collection

    def add_actuator(self, path, unit, actclass, **kwargs):
        setup = kwargs.pop('setup', None)
        a = self.add_timeseries(path, unit, klass=actclass, **kwargs)
        if setup != None:
            a.setup(setup)
        return a

    def set_metadata(self, path, *metadata):
        if len(metadata) > 1:
            metadata = dict([metadata])
        else: metadata = metadata[0]

        for v in metadata.itervalues():
            if not util.is_string(v):
                raise SmapException("set_metadata: values must be strings!")

        o = self.lookup(path)
        o.set_metadata(metadata)

if __name__ == '__main__':
    ROOT_UUID = uuid.uuid1()
    s = SmapInstance(ROOT_UUID)
    s.add_collection("/steve")
    t = Timeseries(s.uuid("sdh"), "V", buffersz=2)
    s.add_timeseries("/sensor0", t)
    s.set_metadata("/sensor0", {
            "Foo" : "Bar",
            "Baz" : 10
            })

    t.add(util.now(), 12)
    t.add(util.now(), 13)
    print s.get_timeseries(t['uuid'])
    print s.get_timeseries('/sensor0')
    print s.get_timeseries('/')

#    s.get_collection('/').set_metadata({'Extra' : {"foo": " bar"}})
    print s.get_collection('/')


#     print "Finding all Timeseries under /"
    print s._lookup_r('/', pred=ITimeseries.providedBy)
    print s.lookup('/+Timeseries')

    print s._lookup_r('/', pred=lambda x: x.dirty)

    # print s._lookup_r("/foo")



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

from zope.interface import Interface, implements

class ITimeseries(Interface):
    """Represent a single time series -- one stream of scalars
    """
    def add(self, *args):
        """Add a new reading to this timeseries.
        
        This version is thread-safe -- can be called from any thread.
        """

    def _add(self, *args):
        """Add a new reading to this timeseries.

        This version should only be called from the twisted main loop.
        """

    def render(self):
        pass

class ITimeseriesFactory(Interface):
    def __call__(unit, 
                 data_type="integer", 
                 timezone="America/Los_Angeles",
                 key=None,
                 new_uuid=None, 
                 buffersz=1):
        """Create a Timeseries

        We set the default data type to Integer and timezone to the
        West Coast, so the only *required* information is the unit of
        measure for this stream.

        If both key and uuid are None, there is an exception because
        one of these is needed to assign the timeseries a UUID.
        Otherwise, "uuid" is used first, followed by a uuid generated
        algorithm 5 based on key + smap.root_uuid.

        @buffersz how many readings are stored in the Reading resource
        (for GET requests, reporting is handled separately).
        """
        print "__CALL__"

class ICollection(Interface):
    """Represent a collection of resources, with attached metadata
    """
    def add_child(self, child):
        """Add a child to the collection"""

    def render(self):
        pass

class IActuator(Interface):
    def setup(self, opts):
        """Set up an actuator"""

    def parse_state(self, state):
        """Parse a state from a string into a state representation"""

    def set_state(self, state):
        """Set the state of the actuator.  The value passed in as the
        new state will be the result of calling parse_state"""

    def get_state(self, state):
        """Read the current state of the actuator.  This should query
        the device for the state, rather than using a cached value."""

class ISmapInstance(Interface):
    """Represents a sMAP instance -- a whole tree of Collections and Timeseries.
    """
    
    def lookup(self, id, pred=None):
        """Look up a sMAP point relative by path or uuid.

        Thread-safe version
        """
    def get_timeseries(self, id):
        """Get a timeseries by identifier"""
    def get_collection(self, id):
        """Get a collection by identifier"""

    def add_timeseries(self, path, *args, **kwargs):
        pass
    def add_collection(self, path, collection):
        pass

class IReporting(Interface):
    """Represents a reporting object, which is responsible for sending
    data to interested consumers.
    """

    def add_report(self, rpt):
        pass
    def get_report(self, rpt):
        pass
    def update_report(self, rpt):
        pass

    def update_subscriptions(self):
        pass

    def publish(self):
        pass


    def _flush(self, force=False):
        """Try to send out all data which is able to be sent currently

        @force ignore MinPeriod and MaxPeriod requests in the reporting instance
        @return a deferred which will fire when all the data has
             delivered.  Any errors in delivery will cause the errBack
             to fire instead.
        """

    def flush(self):
        pass

class ISmapDriver(Interface):
    """Represent a sMAP driver -- something which attaches itself at
    some point in the tree and then may produce readings
    """
    def setup(self, instance, opts={}):
        """The driver should register its points with the relative
        sMAP instance in this call.

        @instance an object implementing ISmapInstanve
        @opts dict of other options to be passed to the driver, ie from a config file
        """

    def start(self):
        """Called when the instance should start producing readings.        
        """



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""
"""Classes for supporting performing actuation with sMAP.

sMAP services with actuation components should place actuators at the
leaves of their resource tree, replacing SmapPoint instances.
Generally, the steps necessary to provide actuation is simply to
subclass the appropriate actuator class (Binary, NState, or
Continuous), and then implement the get_state and set_state methods.

Services wishing to use access control should additionally use SSL for
authentication, and annotate their get and set methods with the
appropriate capabilities necessary to access those resources.
"""
import time
import urlparse
import core
import util
import traceback

from twisted.internet import defer
from zope.interface import implements
from interface import *

class RateLimiter:
    """Class for rate-limiting function call

    @ratelimit minimum time between method calls.  If None,
      method_if_allowed will never be called.  If callable, will allow
      the call if it returns True.
    @method_if_allowed method to be called if allowed by the rate limit
    @method_if_disallowed 
    @return (boolean, result) the boolean indicates if the call was
      allowed by the rate limit, and the result is the result of calling
      the appropriate method

    Since this class overwrites __call__, you can get the result of
    calling the appropriate method just by "calling" this class.it 
    """
    def __init__(self, ratelimit, method_if_allowed=None, method_if_disallowed=None):
        self.ratelimit = ratelimit
        self.method_if_allowed = method_if_allowed
        self.method_if_disallowed = method_if_disallowed
        self.last_value = None
        self.last_call = 0

    def __call__(self, *args, **kwargs):
        allowed = False
        rv = None
        now = time.time()
        if (callable(self.ratelimit) and self.ratelimit()) or \
                (util.is_integer(self.ratelimit) and \
                     now - self.last_call > self.ratelimit):
            allowed = True
            if self.method_if_allowed:
                rv = self.method_if_allowed(*args, **kwargs)
            self.last_call = now
        elif self.method_if_disallowed:
            rv = self.method_if_disallowed(*args, **kwargs)

        return allowed, rv

class SmapActuator(core.Timeseries):
    """Base class for actuators, which deals with HTTP and timing
    requirements.  You must implement three functions for each
    actuator: parse_state, get_state, and set_state.  For common
    actuator classes, parse_state is already done.

    get_state and set_state should be stub functions which talk
    directly to the device; if you wish to use cached value to limit
    the device request rate, you should set read_limit appropriately.

    @unit the value for the UnitofMeasure sMAP field
    @read_limit a value, in seconds, of how often get_state may be
       called.  if not None, the class will call get_state at most once per
       given interval, and otherwise return the last read value that was 
       recorded by calling add().  if None, get_state will never be called and 
       will always return the state recored by calling add().
    @write_limit the same as read_limit, but for set_state.  If
       set_state is called more frequently than this limit, the server
       will return an HTTP error code.
    """
    implements(IActuator)

    def __init__(self, new_uuid, unit, read_limit=0, write_limit=0, autoadd=True, **tsargs):
        core.Timeseries.__init__(self, new_uuid, unit, **tsargs)
        self.FIELDS.append("Actuate")
        self.reader = RateLimiter(read_limit, 
                                  lambda req: util.syncMaybeDeferred(self.get_state, req), 
                                  lambda req: core.Timeseries.render(self, req))
        self.writer = RateLimiter(write_limit, 
                                  lambda req, state: util.syncMaybeDeferred(self.set_state, req, state))
        self.autoadd = autoadd
        self.__setitem__('Actuate', {
                'Model' : self.ACTUATE_MODEL
                })

    def valid_state(self, state):
        """Determine if a given state is valid for a particular actuator.

        The state will be the result of calling parse_state.
        @prototype
        """
        return False

    def parse_state(self, state):
        """Parse a string representation of a state
        @prototype
        """
        return state

    def render(self, request):
        if request.method == 'GET':
            return self.render_read(request)
        elif request.method == 'PUT':
            return self.render_write(request)

    def render_write(self, request):
        """Render a request to change the state"""
        if 'state' in request.args and len(request.args['state']) > 0:
            new_state = self.parse_state(request.args['state'][0])
            if not self.valid_state(new_state):
                raise core.SmapException("Invalid state: " + str(new_state), 400)

            allowed, d = self.writer(request, new_state)
            if allowed:
                d.addCallback(lambda x: self._accept_result_render(request, x))
                return d
            else:
                raise core.SmapException("Cannot actuate now due to rate limit", 503)

    def render_read(self, request):
        """Render the read

        The rate limiter will make sure that we don't overload the
        device; it will used the Timeseries cached value if we've
        called it too much.
        """
        allowed, rv = self.reader(request)
        if allowed:
            rv.addCallback(lambda x: self._accept_result_render(request, x))
        return rv

    def _accept_result_render(self, request, state):
        # now start a render of the underlying TS object
        d = defer.maybeDeferred(core.Timeseries.render, self, request)
        d.addCallback(lambda x: self._finish_render(state, x))
        return d

    def _finish_render(self, state, tsrender):
        # finish by adding the current state as the reading
        now = util.now()

        if self.autoadd:
            self.add(now, state)

        tsrender = dict(tsrender)
        tsrender['Readings'] = [(util.now() * 1000, state)]
        return tsrender


class BinaryActuator(SmapActuator):
    """A BinaryActuator is a controller which has only two states,
generally "on" and "off".  It is essentially a simplified version of
an NStateActuator.

State here are static and can't be configured.
    """
    ACTUATE_MODEL = 'binary'

    def valid_state(self, state):
        return state == 0 or state == 1

    def parse_state(self, state):
        for s in self.control_description['States']:
            if state.strip().lower() in s:
                return int(s[0])
        return None

    def setup(self, opts):
        self.control_type = 'binary'
        self.control_description = {
            'States' : [['0', 'off'], ['1', 'on']]
            }


class NStateActuator(SmapActuator):
    """NStateActuators have a discrete number of states which they can be
in.  Although there may be restrictions on which state transisitions
are possible, this profile does not express any of them.
    """
    ACTUATE_MODEL = 'discrete'
    def valid_state(self, state):
        return int(state) in self.control_description['States']
    
    def __init__(self, states, *args, **kwargs):
        self.control_type = 'nstate'
        self.control_description = {
            'States' : states,
            }
        SmapActuator.__init__(self, **kwargs)

class IntegerActuator(SmapActuator):
    ACTUATE_MODEL = 'integer'
    def valid_state(self, state):
        try:
            int(state)
            return True
        except:
            return False
    
    def parse_state(self, state):
        return int(state)

    def __init__(self, *args, **kwargs):
        self.control_type = 'integer'
        self.control_description = {}
        SmapActuator.__init__(self, *args, **kwargs)

class ContinuousActuator(SmapActuator):
    """A ContinuousActuator allows a set point to be adjusted within a
continuous interval.  Minimum and maximum values in the range must be
specified.
    """
    ACTUATE_MODEL = 'continuous'
    def valid_state(self, state):
        return state >= self.control_description['States'][0] and \
            state <= self.control_description['States'][1]

    def parse_state(self, state):
        return float(state)

    def __init__(self, range=None, **kwargs):
        self.control_type = 'continuous'
        self.control_description = {
            'States' : range,
            }
        SmapActuator.__init__(self, **kwargs)


class GuardBandActuator(SmapActuator):
    """A GuardBandActuator actually consists of two points -- "high" and
"low", which are adjusted in parallel.
    """
    def __init__(self, **kwargs):
        self.control_type = 'guardband'
        SmapActuator.__init__(self, **kwargs)


if __name__ == '__main__':
    import uuid
    import server
    import json
    import sys
    from twisted.python import log
    from authentication import authenticated
    log.startLogging(sys.stdout)
    inst = core.SmapInstance('f80d0504-f2c6-11e0-80e6-ebc97648cfa4')


    class MyActuator(BinaryActuator):
        def setup(self, opts):
            self.state = 0
            BinaryActuator.setup(self, opts)

        def get_state(self, request):
            print request
            print "getting"
            self.add(self.state)
            return self.state
        
        @authenticated(["__has_ssl__"])
        def set_state(self, request, state):
            print "Setting state to", request,state
            self.state = state
            return self.state

#     class MyOtherActuator(ContinuousActuator):
#         def get_state(self, request):
#             return self.state
#         def set_state(self, request, state):
#             print "Setting state to", state
#             self.state = state
    act = MyActuator(inst.uuid('/a1'), 'UoM')
    import actuate

    inst.add_timeseries('/a1', act)
    inst.add_timeseries('/t1', 'V')
    rl = RateLimiter(10)
    
    a2 = inst.add_actuator('/a2', 'UoM', MyActuator, 
                           read_limit=10,
                           write_limit=10,
                           setup={})
    server.run(inst, port=8080)
#     a = MyActuator()
#     b = MyOtherActuator(range=[0, 5])
#     SmapHttp.start_server({'a': a, 'b': b}, port=8000, handler=SmapHttp.SslSmapHandler)



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import time
import operator

from data import escape_string
from smap import core

def build_authcheck(request, ti='', forceprivate=False):
    """Build an SQL WHERE clause which enforces access restrictions.
    Will pull any credentials out of the request object passed in.
    """
    if not 'private' in request.args and not forceprivate:
        query = "(sub%s.public " % ti
    else:
        query = "(false "
    if 'key' in request.args:
        query += 'OR ' + ' OR '.join(["sub.key = %s" % escape_string(x + ti)
                                      for x in request.args['key']])
    query += ")"
    return query

def build_clause(request, clause, ti=''):
    """Build an "inner query" -- a query which yields a list of stream
    ids (indexes in the stream table).  These match the identifiers
    used in the reading db, or can be used as part of a join.  The
    query performs auth checks and will check for the tags specified.
    """
    # the inner query builds a list of streams matching all the
    # clauses which we can then select from

    # perform the auth check in the inner query to give us the
    # most selectivity and avoid returning rows which the user
    # can't access.
    inner_query = ("""
(SELECT stream_id AS cnt
FROM metadata2 mi%(ti)s, subscription sub%(ti)s, stream si%(ti)s
WHERE (%(clause)s) AND """ + build_authcheck(request, ti) + """ AND
   mi%(ti)s.stream_id = si%(ti)s.id AND si%(ti)s.subscription_id = sub%(ti)s.id)
""") % {'clause' : clause, 'ti' : ti}
    return inner_query

class QueryException(core.SmapException):
    pass

class Operator(object):
    """Represent a boolean operator: not, or, and

    Operators have a "render" method which produces SQL
    """
    def __init__(self, clauses=None):
        if not clauses:
            self.clauses = set()
        elif hasattr(clauses, '__iter__'):
            self.clauses = set(clauses)
        else:
            self.clauses = set([clauses])


class Clause(str):
    """A clause is an atom in a a boolean sentence.  In order for
    rendering to work, these must be sql which evaluates to sets of
    streamids"""
    def render(self):
        return self


class AndOperator(Operator):
    """ "and" a set of clauses together
    """
    def check_not(self):
        or_exps = [x.check_not() for x in self.clauses if type(x) == OrOperator]
        and_exps = [x.check_not() for x in self.clauses if type(x) == AndOperator]
        not_exps = [x for x in self.clauses if type(x) == NotOperator]
        clause_exps = [x for x in self.clauses if type(x) == Clause]

        new_exp = set(or_exps + and_exps + not_exps + clause_exps)

        return AndOperator(new_exp)
        
    def render(self):
        not_clauses = filter(lambda x: type(x) == NotOperator, self.clauses)
        and_clauses = self.clauses - set(not_clauses)
        not_transform = OrOperator(map(lambda x: iter(x.clauses).next(), not_clauses))

        rv = ['( (']
        rv.append(' INTERSECT '.join(map(lambda x: x.render(), and_clauses)))
        rv.append(')')
        if len(not_clauses) > 0:
            rv.append('EXCEPT')
            rv.append(not_transform.render())
        rv.append(')')
        return ' '.join(rv)

    def __str__(self):
        return '(' + ' AND '.join(map(str, self.clauses)) + ')'

class OrOperator(Operator):
    """The or operator is a bit complicated because if there are
    "not"ed subclauses, we can't just union them together.  We use
    demorgans rule and convert them into INTERSECT/EXCEPT and then
    negate the result.
    """
    def __str__(self):
        return '(' + ' OR '.join(map(str, self.clauses)) + ')'

    def check_not(self):
        not_clauses = filter(lambda x: type(x) == NotOperator, self.clauses)
        if len(not_clauses) > 0:
            # if there are any negations, group them and use the
            # INTERSECT/EXCEPT method using de morgans rule

            # negate the bare expressions
            or_clauses = map(NotOperator, self.clauses - set(not_clauses))
            # and remove the negation from the negated ones
            not_clauses = map(lambda x: iter(x.clauses).next(), not_clauses)
            # then and the result
            return NotOperator(AndOperator(or_clauses + not_clauses))
        else:
            return self

    def render(self):
        not_clauses = filter(lambda x: type(x) == NotOperator, self.clauses)
        if len(not_clauses) > 0:
            # if there are any negations, group them and use the
            # INTERSECT/EXCEPT method using de morgans rule

            # negate the bare expressions
            or_clauses = map(NotOperator, self.clauses - set(not_clauses))
            # and remove the negation from the negated ones
            not_clauses = map(lambda x: iter(x.clauses).next(), not_clauses)
            # then and the result
            return NotOperator(AndOperator(or_clauses + not_clauses)).render()
        else:
            rv = ['(']
            rv.append(' UNION '.join(map(lambda x: x.render(), self.clauses)))
            rv.append(')')
            return ' '.join(rv)

class NotOperator(Operator):
    """We can't actually render a not operator without implicitly
    ANDing it with the entire database.  For now, we just raise an
    exception.  Nots are actually implemented in "and" using the
    EXCEPT sql operator.
    """
    def __str__(self):
        return "~(" + str(iter(self.clauses).next()) + ')'

    def render(self):
        raise QueryException()

def normalize(stmt):
    """Apply demorgans rule to the query.
    """
    if hasattr(stmt, 'check_not'):
        newstmt = stmt.check_not()
    else:
        newstmt = stmt
    return newstmt

if __name__ == '__main__':
    ast = AndOperator([Clause('c2'), OrOperator([Clause('c1'), NotOperator(Clause('c3'))])])
    print ast
    print normalize(ast).render()




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import os, sys
import uuid
import ConfigParser

try:
    import ordereddict
except ImportError:
    import collections as ordereddict

import core
import util
import driver
import smapconf

class SmapLoadError(core.SmapException):
    """An error was encountered loading a config file"""

def _save_path(conf, inst, path):
    conf.add_section(path)
    cur = inst.lookup(path)

    if not path in inst.drivers:
        conf.set(path, "type", cur.__class__.__name__)
    else:
        # if it's a driver, save the class name that it's loaded as
        conf.set(path, "type", inst.drivers[path].driver.__class__.__module__ + \
                     '.' + inst.drivers[path].driver.__class__.__name__)

    if hasattr(cur, "key"):
        conf.set(path, "key", getattr(cur, "key"))
    elif 'uuid' in cur:
        conf.set(path, "uuid", cur['uuid'])

    if conf.get(path, "type") == "Timeseries":
        conf.set(path, "BufferSize", str(cur['Readings'].size))

    for k, v in cur.iteritems():
        if k in ['uuid', 'Readings', 'Proxy', 'Contents']: continue
        for (name, value) in  util.buildkv(k, v):
            conf.set(path, name, value)

def dump(inst, file):
    """Dump an existing :py:class:`~smap.core.SmapInstance` object to a conf file

:param smap.Core.SmapInstance inst: the object to dump
:param string file: config filename
:raises IOError: if writing to the file fails
"""
    conf = ConfigParser.ConfigParser('', ordereddict.OrderedDict)
    conf.optionxform = str

    q = ['/']
    while len(q) > 0:
        cur = inst.get_collection(q[0])
        if cur and cur.has_key('Contents') and not q[0] in inst.drivers:
            for child in cur['Contents']:
                q.append(util.norm_path(q[0] + '/' + child))
        _save_path(conf, inst, q[0])

        if conf.get(q[0], 'type') == 'Timeseries':
            for k, v in core.Timeseries.DEFAULTS.iteritems():
                if conf.has_option(q[0], k) and \
                        conf.get(q[0], k) == str(v):
                    conf.remove_option(q[0], k)

        q.pop(0)

    with open(file, 'w') as fp:
        conf.write(fp)

def load(file, sections=[], **instargs):
    """Create a sMAP instance based on the representation stored in a file.

The configuration file contains sections which refer to either
reporting instances, or paths in the sMAP heirarchy.  Any section
whose name starts with ``/`` is treated as a resource name; sections
starting with ``report`` are treated as reports.

The file must contain at least one section named ``/``, which must
contain a ``uuid`` key to set the root identifier for the source.

:param string file: filename of the configuration file
:param instargs: arguments passed to the :py:class:`~smap.core.SmapInstance` constructor.
:return smap.core.SmapInstance: the created instancev
:raise smap.loader.SmapLoadError: an error is encountered processing the file
:raise smap.core.SmapError: some other error is encountered validating the loaded object 
    """
    found = None
    for l in ['', os.getcwd(), sys.prefix]:
      path = os.path.join(l, file)
      if os.path.isfile(path):
        found = path
    if not found:
      raise Exception("Config file %s not found." % file)
    print "Loading config file:", found

    conf = ConfigParser.ConfigParser('', ordereddict.OrderedDict)
    conf.optionxform = str
    conf.readfp(open(found, 'r'))

    # if there's a server section, override the default server
    # configuration with that
    if conf.has_section('server'):
        server_conf = dict([(k.lower(), v) for (k,v) in conf.items('server')])
        smapconf.SERVER = util.dict_merge(smapconf.SERVER, server_conf)

    # we need the root to have a uuid
    inst = core.SmapInstance(conf.get('/', 'uuid'), **instargs)
    inst.loading = True
    reports = []

    for s in conf.sections():
        print "Loading section", s
        if s.startswith('report'):
            if conf.has_option(s, 'ReportResource'):
                resource = conf.get(s, 'ReportResource')
            else:
                resource = '/+'
            dest = [conf.get(s, 'ReportDeliveryLocation')]
            for i in xrange(0, 10):
                if conf.has_option(s, 'ReportDeliveryLocation%i' % i):
                    dest.append(conf.get(s, 'ReportDeliveryLocation%i' % i))

            reportinst = {
                'ReportDeliveryLocation' : dest,
                'ReportResource' : resource,
                'uuid' : inst.uuid(s),
                }
            for o in ['MinPeriod', 'MaxPeriod']:
                if conf.has_option(s, o):
                    reportinst[o] = conf.getint(s, o)
            reports.append(reportinst)
            continue
        elif not s.startswith('/'):
            # path sections must start with a '/'
            # other sections might be present and could be parsed by
            # other parts of the program
            print "Warning: skipping section", s, "since it does not begin with a '/'"
            continue
        elif len(sections) and not util.norm_path(s) in sections: 
            # skip all but the listed sections if we were asked to
            continue

        s = util.norm_path(s)

        # build the UUID for the item
        props = util.build_recursive(dict(conf.items(s)))
        id = None
        if conf.has_option(s, 'uuid'):
            key = None
            id = uuid.UUID(conf.get(s, 'uuid'))
        elif conf.has_option(s, 'key'):
            key = conf.get(s, 'key')
        else:
            # default to the path if 
            key = s
        if key:
            id = inst.uuid(key)
            # raise SmapLoadError("Every config file section must have a uuid or a key!")

        # create the timeseries or collection
        if s == '/' or \
               (conf.has_option(s, 'type') and conf.get(s, "type") == 'Collection') or \
               inst.get_collection(s) != None:
            if s == '/':
                c = inst.get_collection('/')
            elif inst.get_collection(s) != None:
                # sometimes you will have collections created twice,
                # for instance if a driver creates it and then we want
                # to tag it with metadata
                c = inst.get_collection(s)
            else:
                c = core.Collection(s, inst)
                inst.add_collection(s, c)
        elif not conf.has_option(s, 'type') or conf.get(s, "type") == "Timeseries":
            if inst.get_timeseries(s) != None:
                c = inst.get_timeseries(s)
            else:   
                try:
                    props['Properties']['UnitofMeasure']
                except KeyError:
                    raise SmapLoadError("A Timeseries must have at least the Properites/UnitofMeasure key")
                
                # the Timeseries uses defaults if the conf file doesn't
                # contain the right sections.
                c = core.Timeseries(id, props['Properties']['UnitofMeasure'],
                                    data_type=props['Properties'].get('ReadingType', 
                                                                      core.Timeseries.DEFAULTS['Properties/ReadingType']),
                                    timezone=props['Properties'].get('Timezone', 
                                                                     core.Timeseries.DEFAULTS['Properties/Timezone']),
                                    buffersz=int(props.get('BufferSize', core.Timeseries.DEFAULTS['BufferSize'])))
                inst.add_timeseries(s, c)
        else:
            if not id:
                raise SmapLoadError("A driver must have a key or uuid to generate a namespace")
            
            # load a new driver manager layer
            newdrv = driver.SmapDriver.get_driver(inst, conf.get(s, 'type'), s, id)
            # create a collection and add it at the attachment point
            c = inst.get_collection(s)
            if not c:
                c = core.Collection(s, inst)
                inst.add_collection(s, c)

            # get the driver to add its points
            newdrv.setup(dict(conf.items(s)))

        # Metadata and Description are shared between both Collections
        # and Timeseries
        if props.has_key('Metadata'):
            # the driver may have added metadata; however config file
            # metadata overrides it
            c['Metadata'] = util.dict_merge(c.get('Metadata', {}),
                                            props['Metadata'])
        if props.has_key('Description'):
            c['Description'] = props['Description']
        if key:
            setattr(c, 'key', key)

    # since the sections could come in any order, update the reporting
    # instance to make sure all the topics are set right.
    for reportinst in reports:
        if not inst.reports.update_report(reportinst):
            inst.reports.add_report(reportinst)
    inst.reports.update_subscriptions()
    inst.loading = False
    return inst



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import os
import sys
import time
import traceback

import util
import json
import copy
import pprint
import cStringIO as StringIO

from twisted.internet import reactor, task, defer, threads
from twisted.web.client import Agent
from twisted.web.http_headers import Headers
from twisted.python import log
import logging

import core
import disklog
from contrib import client

# this is the largest number of records we will store.
BUFSIZE_LIMIT = 100000

# the most records to pack into a single log entry/message to the server
REPORT_RECORD_LIMIT = 10000

def reporting_copy(obj):
    if isinstance(obj, dict):
        rv = dict(obj)
        for k in rv.keys():
            rv[k] = reporting_copy(rv[k])
        return rv
    elif isinstance(obj, list):
        rv = list(obj)
        for i in xrange(0, len(rv)):
            rv[i] = reporting_copy(rv[i])
        return rv
    else:
        return obj

"""Iterate over all the colletions and timeseries contained in a
report object.
"""
def reporting_map(rpt, col_cb, ts_cb):
    q = ['/']
    while len(q) > 0:
        cur_path = util.norm_path(q.pop(0))
        cur = rpt.get(cur_path, None)
        if not cur: continue
        if 'Contents' in cur:
            for pc in cur['Contents']:
                q.append(cur_path + '/' + pc)
            col_cb(cur_path, cur)
        else:
            ts_cb(cur_path, cur)
        del rpt[cur_path]

    for p, v in rpt.iteritems():
        if 'Contents' in v:
            col_cb(p, v)
        else:
            ts_cb(p, v)

"""Push all metadata down to the leaves and remove the collections
"""

def push_metadata(rpt):
    for k, v in rpt.iteritems():
        sp = util.split_path(k)
        if 'Readings' in v:
            for i in xrange(0, len(sp)):
                if util.join_path(sp[:i]) in rpt:
                    upobj = rpt[util.join_path(sp[:i])]
                    if 'Contents' in upobj:
                        del upobj['Contents']
                    v.update(util.dict_merge(upobj, v))
    for k, v in rpt.items():
        if not 'Readings' in v:
            del rpt[k]

class DataBuffer:
    """Buffer outgoing data.

    This buffer class allows you to add readings which are stored into
    a circular buffer.  It supports split-phase adding and delivery by
    returning a "truncation specification" along with the data so that
    if you later wish to shorten the buffer to remove previously read
    data, you can do that.  Note that since this is a circular buffer,
    that may mean removing less data than you read back (some may have
    been overwritten).
    """
    def __init__(self, datadir):
        """
        :param int max_size: the maximum total log size, in records
        """
        log.msg("Create databuffer " + datadir)
        self.datadir = datadir
        self.data = disklog.DiskLog(datadir)
        self.head_metric = self.metric(self.data.tail())
        # if we do a read, we have to add a new log record so we don't
        # keep appending to the same one
        self.clear = False

    def __str__(self):
        return "DataBuffer len: %i" % len(self.data)

    def __len__(self):
        return len(self.data)

    def sync(self):
        self.data.sync()

    def metric(self, val):
        if hasattr(val, '__iter__'):
            if 'Readings' in val:
                val_metric = len(val['Readings'])
            else: val_metric = sum(map(len, val))
        else:
            val_metric = 1
        return val_metric

    def add(self, key, val):
        """Enqueue a new object for delivery with a subscription

        :param string key: The key for the data stream
        :param string val: The new value for the object.  Copied.
        """
        tail = self.data.tail()

        # length of the head
        val_metric = self.metric(val)

        if tail == None or self.clear:
            self.data.append({key: reporting_copy(val)})
            self.head_metric = val_metric
            self.clear = False
        elif 'Contents' in val and len(val['Contents']) == 0:
            # okay to skip b/c it doesn't apply to anything
            pass
        elif key in tail and len(val) == 2 and \
                'Readings' in val and 'uuid' in val:
            if not 'Readings' in tail[key]:
                tail[key]['Readings'] = []
            if self.head_metric < REPORT_RECORD_LIMIT:
                tail[key]['Readings'].extend(val['Readings'])
                self.data.update_tail(tail)
                self.head_metric += val_metric
            else:
                self.data.append({key: reporting_copy(val)})
                self.head_metric = val_metric

        # really this might just want to merge updates...
        elif key in tail and val == tail[key]:
            pass
        elif not key in tail and self.head_metric < REPORT_RECORD_LIMIT:
            tail[key] = reporting_copy(val)
            self.data.update_tail(tail)
            self.head_metric += val_metric
        else:
            self.data.append({key: reporting_copy(val)})
            self.head_metric = val_metric

    def truncate(self):
        """Truncate a set of readings based on the sequence number
        stored from a previous read"""
        self.data.pop()
        
    def read(self):
        """Read n points (per stream) back as a flat list.  AlsoFi
        return a truncation specification so we can remove this data
        later if desired

        :rvalue: readings, tspec.  ``readings`` is the reading object
        you can send to a consumer.  ``tspec`` can be passed to
        ``truncate()`` and will removme exactly the data you just read
        -- note this is necessary because the log is a circular buffer
        and may have wrapped while you were processing the readings.
        """
        if len(self.data) > 0:
            self.clear = True
            return self.data.head()
        else:
            raise core.SmapException("No Pending Data!")


class ReportInstance(dict):
    """Represent the stored state pending for one report destination
    """
    def __init__(self, datadir, *args):
        dict.__init__(self, *args)
        if not 'MinPeriod' in self:
            self['MinPeriod'] = 0
        if not 'MaxPeriod' in self:
            self['MaxPeriod'] = 2 ** 31 - 1
        self['DataDir'] = datadir
        self['PendingData'] = DataBuffer(datadir)
        self['ReportDeliveryIdx'] = 0
        self['LastAttempt'] = 0
        self['LastSuccess'] = 0
        self['Busy'] = False

    def deliverable(self):
        """Check if attempt should be called
        :rvalue boolean: True if a report should be sent
        """
        now = util.now()
        return (now - self['LastSuccess'] > self['MaxPeriod']) or \
            (len(self['PendingData']) > 0 and \
                 (now - self['LastSuccess']) > self['MinPeriod'])

    def attempt(self):
        """Try to make a delivery
        :rvalue: a :py:class:`Deferred` of the attempt
        """
        if 'Busy' in self and self['Busy']:
            return

        try:
            data = self['PendingData'].read()
            if data == None:
                self['PendingData'].truncate()
                return
        except Exception, e:
            print e
            return

        self['LastAttempt'] = util.now()
        log.msg("publishing to %s: %i %s" % 
                (self['ReportDeliveryLocation'][self['ReportDeliveryIdx']],
                 len(data), 
                 str([len(x['Readings']) 
                      for x in data.itervalues() 
                      if 'Readings' in x])),
                logLevel=logging.DEBUG)
        # set up an agent to push the data to the consumer
        agent = Agent(reactor)
        try:
            v = StringIO.StringIO()
            util.dump_json(data, v)
            v.seek(0)
            d = agent.request('POST',
                              str(self['ReportDeliveryLocation'][self['ReportDeliveryIdx']]),
                              Headers({'Content-type' : 
                                       ['application/json']}),
                              client.FileBodyProducer(v))
        except:
            traceback.print_exc()
            return

        def makeSuccessCb():
            # make a closure for removing the delivered data
            # on a success
            def cbResponse(resp):
                self['Busy'] = False
                if resp.code in [200, 201, 204]:
                    # on success record the time and remove the data
                    self['LastSuccess'] = util.now()
                    self['PendingData'].truncate()
                    if len(self['PendingData']) > 0:
                        # this causes a new deferred to get added to
                        # the chain, so we continue immediately at the
                        # next log position if this one worked.
                        return self.attempt()
                    else:
                        return resp
                else:
                    # but most HTTP codes indicate a failure
                    log.err("Report delivery to %s returned %i" % (
                            self['ReportDeliveryLocation'][self['ReportDeliveryIdx']],
                            resp.code))
                    self['ReportDeliveryIdx'] = ((self['ReportDeliveryIdx'] + 1) %
                                                 len(self['ReportDeliveryLocation']))
                    raise core.SmapException("Report delivery to " +
                                             str(self['ReportDeliveryLocation']) +
                                             ' returned ' + str(resp.code))
            return cbResponse

        def makeDoneCb(inst):
            inst_ = inst
            def doneCb(resp):
                inst_['Busy'] = False
                self['ReportDeliveryIdx'] = ((self['ReportDeliveryIdx'] + 1) %
                                             len(self['ReportDeliveryLocation']))
                return resp
            return doneCb

        self['Busy'] = True
        d.addErrback(makeDoneCb(self))
        d.addCallback(makeSuccessCb())
        return d


class Reporting:
    def __init__(self, inst, autoflush=1.0, reportfile=None, max_size=BUFSIZE_LIMIT):
        """Create a new report manager, responsible for delivering
        data to subscribers.  Buffers data on disk until it can be
        delivered, storing up to *max_size* points per timeseries.

        :param inst: the :py:class:`~smap.core.SmapInstance` we'll be
         delivering reports for.
        :param float autoflush: how often to call flush, which attempts
         to deliver all data.
        :param string reportfile: backing store for reporting instances
         and data which hasn't been delivered yet.
        :param int max_size: the maximum number of points we will buffer for 
         any stream.
        """
        self.inst = inst
        self.subscribers = []
        self.reportfile = reportfile
        self.max_size = max_size
        self.autoflush = autoflush
        if self.reportfile:
            self.load_reports()

        if autoflush != None:
            # add a blank callback to the deferred to kill off any errors
            def flush():
                d = self._flush()
                if d: d.addBoth(lambda _ : None)
            self.t = task.LoopingCall(flush)
            self.t.start(autoflush)

        # add a shutdown handler so we save the final reports after exiting
        if reportfile:
            reactor.addSystemEventTrigger('after', 'shutdown', 
                                          self.save_reports)

    def get_report(self, id):
        return util.find(lambda item: item['uuid'] == id, self.subscribers)

    def add_report(self, rpt):
        dir = os.path.join(self.reportfile + '-reports',
                           str(rpt['uuid']))
        report_instance = ReportInstance(dir, rpt)
        log.msg("Creating report -- dest is %s" % str(rpt['ReportDeliveryLocation']))
        self._update_subscriptions(report_instance)
        self.subscribers.append(report_instance)

        # publish the full data set when we add a subscription so we
        # can compress from here
        for k, v in self.inst.lookup(report_instance['ReportResource']).iteritems():
            self.publish(k, v)

    def del_report(self, id):
        rpt = self.get_report(id)
        if rpt:
            log.msg("removing report " + str(id))
            del self.subscribers[self.subscribers.index(rpt)]
        return rpt

    def update_report(self, rpt):
        """Copy in fields from a new reporting object
        """
        cur = self.get_report(rpt['uuid'])
        log.msg("updating report %s: dest is now %s" % (
                str(rpt['uuid']),
                str(rpt['ReportDeliveryLocation'])))
        if cur:
            cur.update(rpt)
            self._update_subscriptions(cur)
            return True
        return False

    def _update_subscriptions(self, sub):
        result = self.inst.lookup(sub['ReportResource'])
        if isinstance(result, dict):
            sub['Topics'] = set(result.iterkeys())
        elif isinstance(result, dict) and 'uuid' in result:
            sub['Topics'] = set([getattr(result, 'path')])
        else:
            sub['Topics'] = set()

    def update_subscriptions(self):
        """Should be called whenever the set of resources changes so we can
        update the list of uuids for each subscriber."""
        map(self._update_subscriptions, self.subscribers)

    def publish(self, path, val, prepend=False):
        """Publish a new reading to the stream identified by a path.

        Not thread safe.
        """
        path = util.norm_path(path)
        for sub in self.subscribers:
            if path in sub['Topics']:
                sub['PendingData'].add(path, val)

    def load_reports(self):
        self.subscribers = util.pickle_load(self.reportfile)
        if self.subscribers == None:
            self.subscribers = []
        for s in self.subscribers:
            s['Busy'] = False

    def save_reports(self, *args):
        """Save reports while holding the filesystem lock.
        """
        util.pickle_dump(self.reportfile, self.subscribers)
        for s in self.subscribers:
            s['PendingData'].sync()

        if len(args) == 1:
            return args[0]
        else:
            return 

    def _flush(self, force=False):
        """Send out json-packed report objects to registered listeners. 
        
        :param boolean force: if True, ignore ``MinPeriod``/``MaxPeriod``.
        :rtype: a :py:class:`twisted.internet.task.DeferredList`
         instance which will fire when deliver to all subscribers has
         finished, or errBack when any fail
        """
        deferList, deleteList = [], []
        for sub in self.subscribers:
            now = util.now()
            if sub.get('ExpireTime', now) < now:
                # remove expired reports
                deleteList.append(sub['uuid'])

            # either we've gone too long without trying and so need to
            # deliver a report or else we have new data and have
            # waited for at least MinPeriod since the last report.
            elif force or sub.deliverable():
                d = defer.maybeDeferred(sub.attempt)
                # we don't need an errBack for this case since we want
                # to propagate the error and don't need to do any
                # cleanup
                deferList.append(d)

        map(self.del_report, deleteList)
        d = defer.DeferredList(deferList, fireOnOneErrback=True, consumeErrors=True)
        d.addBoth(self.save_reports)
        return d

    def flush(self):
        """Causes delivery to be attempted for all non-busy reports
        now.  Threaded version; blocks until :py:meth:`_flush` finishes.
        """
        return threads.blockingCallFromThread(reactor, self.inst._flush)




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import uuid
import json

import pprint
import traceback

from twisted.internet import reactor, task, defer, threads
from twisted.web.client import Agent
from twisted.web.http_headers import Headers
from twisted.python import log
from twisted.web import resource, server, proxy

import util
import reporting 

class SmapConsumer(resource.Resource):
    def __init__(self):
        self.streams = {}
        resource.Resource.__init__(self)

    def add(self, report):
        try:
            reporting.push_metadata(report)
        except Exception, e:
            traceback.print_exc()
        for path, val in report.iteritems():
            with open('data/' + path.replace('/', '-'), 'a') as fp:
                readings = val.pop('Readings')
                if len(readings) == 0: continue
                val.pop('uuid')
                for k, v in util.buildkv('', val):
                    print >>fp, "# %s = %s" % (k[1:], v)
                for rv in readings:
                    print >>fp, rv['ReadingTime'], rv['Reading']

    isLeaf = True
    def render_POST(self, request):
        print 
        obj = json.load(request.content)
        # pprint.pprint(obj)
        self.add(obj)
        request.finish()
        return server.NOT_DONE_YET
        

class SmapSubscriber:
    """Class to maintain a subscription to a sMAP source.
    """
    def __init__(self, url, dest, resource='/+', id=None, expire_time=60, min_period=0):
        """
        :param string url: The URL of the sMAP server
        :param string dest: The URL of the data destination
        :param uuid.UUID id: id of the reporting instance to touch
        :param string expire_time: how long before the sMAP server
         automatically deletes the report, in seconds
        :param string min_period: reporting min period value -- don't
         deliver reports more frequently than this.
        """
        self.url = url
        self.expire_time = expire_time
        if not id: id = uuid.uuid1()
        self.rpt_obj = {
            'uuid' : str(id),
            'ReportDeliveryLocation' : [dest],
            'ReportResource' : resource,
            'MinPeriod': int(min_period * 1000)
            }
        self.is_subscribed = False

    def is_subscribed(self):
        return self.is_subscribed

    def subscribe(self):
        """Send or re-send the request for data.
        """
        if self.expire_time:
            self.rpt_obj['ExpireTime'] = util.now() + int(self.expire_time * 1000)
        agent = Agent(reactor)
        d = agent.request('PUT',
                          self.url + '/reports/' + str(self.rpt_obj['uuid']),
                          Headers({'Content-type' : ['application/json']}),
                          util.AsyncJSON(self.rpt_obj))
        def eb(request):
            self.is_subscribed = False
            log.err("Subscription failed to " + self.url)
            return False
        def sb(request):
            if not isinstance(request, bool) and request.code in [200, 201]:
                log.msg("Successfully subscribed to " + self.url)
                self.is_subscribed = True
                return True
            else:
                return eb(request)
        d.addErrback(eb)
        d.addCallback(sb)
        return d

    def unsubscribe(self):
        """Delete an installed reporting instance
        """
        agent = Agent(reactor)
        d = agent.request('DELETE',
                          self.url + '/reports/' + str(self.rpt_obj['uuid']),
                          Headers({'Content-type' : ['application/json']}))
        def sb(request):
            if not isinstance(request, bool) and request.code in [200, 201, 404]:
                self.is_subscribed = False
                log.msg("unsubscribed from " + self.url)
                return True
            else: return False
        d.addCallback(sb)
        return d
           

def _subscribe(result, settings):
    """
    Called with list of sMAP sources to subscribes to
    Returns map of uuid => :py:class:`smap.subscriber.SmapSubcriber` instances 
    """
    subs = {}
    for (url, id, key) in result:
        dest = 'http://%s:%s/add/%s' % (settings.MY_LOCATION,
                                         settings.MY_PORT,
                                         key)
        s =SmapSubscriber(url, dest, id=id, expire_time=None)
        s.subscribe()
        subs[id] = s
    return subs

def subscribe(db, settings):
    """Look up all of the sMAP sources we should subscribe to
    """
    d = db.runQuery("""
       SELECT url, uuid, key 
       FROM subscription 
       WHERE url IS NOT NULL AND url != ''""")
    d.addCallback(_subscribe, settings)
    return d





"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import re
import pprint
from BeautifulSoup import BeautifulSoup as bs
from smap.iface.http.httputils import load_html
from optparse import OptionParser
import ConfigParser
import urllib
import urlparse

import ConfigParser
try:
    import ordereddict
except ImportError:
    import collections as ordereddict


try:
    import ordereddict
except ImportError:
    import collections as ordereddict

sys.path.append("../../newlib")
import sensordb
import auth


BMOROOT = 'http://www.buildingmanageronline.com/members/'
STATUSPAGE = 'client_status.php?DB=dbU216ucberkelF682'
AUTH = ('ucbguest', 'ucbguest')

def remove_entities(s):
    return re.sub('&.*;', '', s)

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option('-t', '--types', dest='types', action='store_true', default=False,
                      help='print the types of devices found')
    parser.add_option('-n', '--no-cache', dest='cache', action='store_false', default=True,
                      help='do not use cached pages')
    parser.add_option('-b', '--buildings', dest='buildings', action='store_true', default=False,
                      help='show building list')
    parser.add_option('-p', '--progress', dest='progress', action='store_true', default=False,
                      help='show buildings as we load them')
    parser.add_option('-l', '--load', dest='load', action='store_true', default=False,
                      help='generate conf for loading data from bmo')
    parser.add_option('-d', '--database', dest='db', action='store_true', default=False,
                      help='generate a prototype sensordb for an instrument')
    (opts, args) = parser.parse_args()
    opts.conf = True
    if opts.types or opts.buildings or opts.load or opts.db:
        opts.conf = False

    # find all the AcquiSuite boxes
    devices = {}
    soup = load_html(BMOROOT + STATUSPAGE, auth=AUTH, cache=opts.cache)
    for tr in soup.findAll('tr'):
        tds = tr('td')
        if len(tds) != 6: continue
        
        name = tds[0].a.string
        devices[name] = {
            'ip' : remove_entities(tds[3].string),
            'href' : tds[0].a['href'],
            }

    # look at all the meters hanging off each of them
    for location in devices.iterkeys():
        if opts.progress:
            print >>sys.stderr, "Processing", location
        soup = load_html(BMOROOT + devices[location]['href'], auth=AUTH, cache=opts.cache)
        subdevices = []
        for tr in soup.findAll('tr'):
            tds = tr('td')
            if len(tds) != 5 or tds[3].a != None: continue

            subdevices.append({
                    'address' : remove_entities(tds[0].string),
                    'name' : remove_entities(tds[2].string),
                    'type' : remove_entities(tds[3].string)
                    })
        devices[location]['subdevices'] = subdevices
    
    if opts.types:
        mtypes = {}
        for name, v in devices.iteritems():
            for t in v['subdevices']:
                if not t['type'] in mtypes:
                    mtypes[t['type']] = {'count': 0, 'locs': []}
                mtypes[t['type']]['count'] = mtypes[t['type']]['count'] + 1
                mtypes[t['type']]['locs'].append(name)
        for n, c in mtypes.iteritems():
            print >>sys.stderr, c['count'], n, ' '.join(c['locs'])

    if opts.buildings:
        print >>sys.stderr, '\n'.join(devices.iterkeys())

if opts.conf:
    def make_section(cmps):
        cmps = map(lambda s: s.replace(' ', '_'), cmps)
        return '/' + '/'.join(map(lambda s: urllib.quote(s, safe=''), cmps))

    conf = ConfigParser.ConfigParser('', ordereddict.OrderedDict)
    conf.optionxform = str
    conf.add_section('/')
    conf.set('/', 'Metadata/Location/Campus', 'UCB')
    conf.set('/', 'type', 'Collection')
    
    for location, devs in devices.iteritems():
        if not location in auth.AUTH: continue

        parent_sec = make_section((location, ))
        conf.add_section(parent_sec)
        conf.set(parent_sec, 'type', 'Collection')
        conf.set(parent_sec, 'Metadata/Location/Building', location)
        
        sec = None
        for d in devs['subdevices']:
            map = sensordb.get_map(d['type'], location)
            if map != None:
                sec = make_section((location, d['name']))
                conf.add_section(sec)
                conf.set(sec, 'type', 'smap.drivers.obvius.obvius.Driver')
                conf.set(sec, 'Username', auth.AUTH[location][0])
                conf.set(sec, 'Password', auth.AUTH[location][1])
                conf.set(sec, 'Url', 'http://' + devs['ip'] + 
                         '/setup/devicexml.cgi?ADDRESS=%s&TYPE=DATA' % d['address'],)
                conf.set(sec, 'ObviousType', d['type'])
        if sec == None:
            conf.remove_section(parent_sec)

    conf.write(sys.stderr)
elif opts.load:
    conf = {}
    for location, devs in devices.iteritems():
        params = urlparse.parse_qs(urlparse.urlsplit(devs['href']).query)
        if not "AS" in params or not  "DB" in params: continue
        if location in auth.AUTH: continue
        thisconf = {}
        for d in devs['subdevices']:
            if sensordb.get_map(d['type'], location) != None:
                dlurl = BMOROOT + 'mbdev_export.php/' + params['AS'][0] + '_' +  \
                    d['address'] + '.csv' + "?DB=" + params['DB'][0] + '&AS=' + \
                    params['AS'][0] + '&MB=' + d['address'] + '&DOWNLOAD=YES' + \
                    "&COLNAMES=ON&EXPORTTIMEZONE=UTC&DELIMITER=TAB" + \
                    '&DATE_RANGE_STARTTIME=%s&DATE_RANGE_ENDTIME=%s'
                thisconf[d['name']] = (
                    d['type'],
                    dlurl)

        if len(thisconf) > 0:
            conf[location] = thisconf

    # generate config file
    cf = ConfigParser.ConfigParser('', ordereddict.OrderedDict)
    cf.optionxform = str
    import obvius
    cf.add_section('server')
    cf.set('server', 'SuggestThreadPool', '20')
    cf.set('server', 'Port', '9051')

    cf.add_section('/')
    cf.set('/', 'Metadata/Location/Campus', 'UCB')
    cf.set('/', 'Metadata/SourceName', 'buildingmanageronline archive')
    cf.set('/', 'uuid', '91dde108-d02b-11e0-8542-0026bb56ec92')

    for building in conf.iterkeys():
        building_path = '/' + obvius.to_pathname(building)
        cf.add_section(building_path)
        cf.set(building_path, 'type', 'Collection')
        for metername in conf[building].iterkeys():
            metertype, url = conf[building][metername]

            building_name = building
            if "New" in building_name:
                building_name = building_name[:building_name.index("New")]
            if "NEW" in building_name:
                building_name = building_name[:building_name.index("NEW")]

            meter_path = building_path + '/' + obvius.to_pathname(metername)
            cf.add_section(meter_path)
            cf.set(meter_path, 'type', 'smap.drivers.obvius.bmo.BMOLoader')
            cf.set(meter_path, 'Metadata/Extra/MeterName', metername)
            cf.set(meter_path, 'Metadata/Instrument/Model', metertype)
            cf.set(meter_path, 'Metadata/Location/Building', building_name)
            cf.set(meter_path, 'Url', url)

            # add any extra config options specific to this meter type
            map = sensordb.get_map(metertype, building_name)
            if 'extra' in map:
                for k,v in map['extra'].iteritems():
                    cf.set(meter_path, k, v)


    cf.write(sys.stdout)
#     else:
#         for k, v in conf.iteritems():
#             if len(v) > 1:
#                 for (meter, ) in v.iteritems():
#                     print '/' + obvius.to_pathname(k) + '/' + obvius.to_pathname(meter) + ',' + k + " " + meter
#             else:
#                 print '/' + obvius.to_pathname(k) + '/' + obvius.to_pathname(v.keys()[0]) + ',' + k

elif opts.db:
    print "generate sensordb"
    for location, devs in devices.iteritems():
        print devs




import uuid
import numpy as np
from scipy.stats import nanmean

from smap.core import SmapException
from smap.archiver.client import SmapClient
from smap.operators import *
from smap import util
import smap.operators as o

def _mean(vec):
    """Take the mean of both the timestamps and the values"""
    if len(vec) == 0:
        return null, {}
    else:
        return np.array([np.mean(vec, axis=0)]), {}

def joinedop(op, inputs):
    vals = op(map(lambda x: x[:,1], inputs), axis=0)
    return [np.dstack((inputs[0][:,0], vals))[0]]

joinedsum = lambda input: joinedop(np.nansum, input)
joinedmean = lambda input: joinedop(nanmean, input)
joinedmedian = lambda input: joinedop(np.median, input)

def joinedsub(inputs):
    assert len(inputs) == 2
    vals = np.subtract(*map(lambda x: x[:,1], inputs))
    return [np.dstack((inputs[0][:,0], vals))[0]]
    
##
## Buffering operators. BufferedJoinOperator and LatestOperator are
## often placed at the head of the processing chain, so later
## operators can operate on windows of data.
##
##
## Mean operators with different semantics
##
class _MeanOperator(Operator):
    """Mean operator which requires all data to be present and
    timestamps to be aligned"""
    name = 'mean'
    process = staticmethod(joinedmean)

class _MeanVectorOperator(ParallelSimpleOperator):
    """Take the mean of all data presented from N feeds"""
    name = 'mean vector'
    base_operator = staticmethod(_mean)

class _SumOperator(Operator):
    name = 'sum'
    process = staticmethod(joinedsum)

class _SubtractOperator(Operator):
    name = 'subtract'
    process = staticmethod(joinedsub)

class OrderOperator(Operator):
    def __init__(self, inputs, order_tag):
        self.name = 'order-by-%s' % order_tag
        Operator.__init__(self, inputs, OP_N_TO_N)
        self.reorder = map(operator.itemgetter(0),
            sorted(enumerate(inputs), key=lambda x: x[1][order_tag]))

    def process(self, inputs):
        # apply the operator to the reordered inputs
        return map(lambda x: inputs[x], self.reorder)

class BucketOperator(GroupbyTimeOperator):
    name = 'bucketed mean operator'
    operator_name = 'bucket'
    operator_constructors = [(),
                             (int,),
                             (int, float)]
    def __init__(self, inputs, chunk_length=10, chunk_delay=1):
        GroupbyTimeOperator.__init__(self, inputs,
                                     _MeanVectorOperator, 
                                     chunk_length, chunk_delay)


##
## Subsampling operators
##
##
## Useful operators made up of streams of other operators
##
class SubsampledSumOperator(CompositionOperator):
    """Add up N streams by first bucketing and then buffering data
    until data from all streams are present.
    """
    name = 'subsampled sum'
    def __init__(self, inputs, windowsz):
        self.windowsz = windowsz
        self.oplist = [
            StandardizeUnitsOperator,
            lambda inputs: SubsampleOperator(inputs, self.windowsz),
            BufferedJoinOperator,
            _SumOperator]
        CompositionOperator.__init__(self, inputs)

class MissingSumOperator(CompositionOperator):
    name = "sum with missing data"
    def __init__(self, inputs, windowsz, missingthresh=0.6):
        self.oplist = [
            StandardizeUnitsOperator,
            lambda inputs: GroupbyTimeOperator(inputs, 
                                               lambda x: SubsampleOperator(x, windowsz),
                                               chunk_length=windowsz),
            lambda inputs: o._MissingDataOperator(inputs, missingthresh),
            # PrintOperator,
            _SumOperator,
            ]
        CompositionOperator.__init__(self, inputs)


class SubsampleMeanOperator(CompositionOperator):
    name = 'subsample-mean'
    operator_name = 'subsample'
    operator_constructors = [(int,)]

    def __init__(self, inputs, windowsz=300000):
        self.oplist = [
            # take the mean of each bucket
            lambda inputs: GroupbyTimeOperator(inputs,
                                               _MeanVectorOperator,
                                               chunk_length=windowsz),
            # but snap the timestamps to the beginning of the window
            lambda inputs: SnapTimes(inputs, windowsz),
            ]
        CompositionOperator.__init__(self, inputs)


class MeanOperator(CompositionOperator):
    """Maintain a mean of N streams 
    """
    name = 'Mean'
    operator_name = 'mean'
    operator_constructors = [(int, )]
    def __init__(self, inputs, windowsz=300):
        self.oplist = [
            lambda inputs: GroupbyTimeOperator(inputs,
                                               _MeanVectorOperator,
                                               chunk_length=windowsz),
            
            # then take the mean across all feeds
            _MeanOperator,

            # snap times to window starts
            lambda inputs: SnapTimes(inputs, windowsz),
            ]
        CompositionOperator.__init__(self, inputs)

class SumOperator(CompositionOperator):
    operator_name = 'sum'
    operator_constructors = [(),
                             (int,),
                             (int, float, float)]

    def __init__(self, inputs, windowsz=300, delay=1.0, data_fraction=0.9):
        self.name = 'sum-%i' % windowsz
        self.oplist = [  
            StandardizeUnitsOperator,

            lambda inputs: GroupbyTimeOperator(inputs,
                                               _MeanVectorOperator,
                                               chunk_length=windowsz,
                                               chunk_delay=delay),

            lambda inputs: o._MissingDataOperator(inputs, data_fraction),

            # gotta fix them units before we add
            StandardizeUnitsOperator,

            # then take the mean across all feeds
            _SumOperator,

            # snap times to window starts
            lambda inputs: SnapTimes(inputs, windowsz),
            # PrintOperator,
            ]
        CompositionOperator.__init__(self, inputs)
        

class SubtractOperator(CompositionOperator):
    """Compute the mean of differences of streams
    """
    name = "subtract"
    operator_name = 'subtract'
    operator_constructors = [(str,),
                             (str, int)]

    def __init__(self, inputs, inner_order=None, windowsz=300):
        self.oplist = [
            lambda inputs: GroupbyTimeOperator(inputs, 
                                               _MeanVectorOperator,
                                               chunk_length=windowsz),
            lambda inputs: OrderOperator(inputs, inner_order),
            _SubtractOperator,
            lambda inputs: SnapTimes(inputs, windowsz),
            # PrintOperator,
            ]
        CompositionOperator.__init__(self, inputs)

class WindowedDriver(GroupedOperatorDriver):
    INIT_ARGS = (
        ('Window', 'windowsz', 300, int, True),
        ('Delay', 'delay', None, float, False),
        ('DataFraction', 'data_fraction', None, float, False),
        ('OrderKey', 'inner_order', None, str, False),
        )

    def setup(self, opts):
        kwargs = {}
        for arg in self.INIT_ARGS:
            name, argname, default, cvtr, mandatory = arg
            if mandatory or name in opts:
                val = opts.get(name, default)
                kwargs[argname] = cvtr(val)

        self.operator_class = (lambda x: self.inner_operator(x, **kwargs))
        GroupedOperatorDriver.setup(self, opts)

class DefaultSummationDriver(WindowedDriver):
    inner_operator = SubsampledSumOperator

class MissingSummationDriver(WindowedDriver):
    inner_operator = MissingSumOperator

class SubsampleMeanDriver(WindowedDriver):
    inner_operator = SubsampleMeanOperator

class MeanDriver(WindowedDriver):
    inner_operator = MeanOperator

class SumDriver(WindowedDriver):
    inner_operator = SumOperator

class SubtractDriver(WindowedDriver):
    inner_operator = SubtractOperator

if __name__ == '__main__':
    ip = [{
            'uuid' : '7ebdde28-44a9-11e1-8968-00508dca5a06',
            'tag' : 'foo',
            },
          {
            'uuid' : '7f2fef2c-44a9-11e1-afa5-00508dca5a06',
            'tag' : 'boo',
            }]
    
    # o = OrderedArithmeticOperator(ip, 'tag')
    o = SubtractOperator(ip, 'tag')
    print o([np.array([[0, 1], [600, 2]]), 
             np.array([[0, 2], [600, 3]])])



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
'''
sMAP feed for CA ISO energy market data. Provides price feeds for three markets at a specific location:
day-ahead market (DAM), hour-ahead scheduling process (HASP), real-time market (RTM).
Each price feed provides total price, loss price, energy price and congestion cost.

@author Andrew Krioukov
 conversion to smap-2.0
@author Stephen Dawson-Haggerty
'''

import zipfile
import urllib2
import StringIO
import BeautifulSoup
import datetime

import sys
import logging
import time
import threading
from operator import itemgetter, attrgetter

from zope.interface import implements

from smap.driver import SmapDriver
from smap.util import periodicSequentialCall
from smap.contrib import dtutil

class CaIsoPrice(SmapDriver):

  MARKETS = [('DAM', 30*60, 'Day-ahead market'),
             ('HASP', 10*60, 'Hour-ahead scheduling process'),
             ('RTM', 2*60, 'Real-time market')]
  FEEDS = [('total_price', '$', 'total price'),
           ('loss', '$', 'loss price'),
           ('congestion', '$', 'congestion price'),
           ('energy', '$', 'energy price')]

  def setup(self, opts):
    # get our location
    self.last_reading = {}
    for m, t, d in self.MARKETS:
      self.last_reading[m] = 0

    self.location = opts.get('Location', 'OAKLAND_1_N001')
    self.set_metadata('/', {
      'Location/Uri' : 'http://oasis.caiso.com/mrtu-oasis/SingleZip',
      'Extra/IsoNode' : self.location,
      'Extra/Driver' : 'smap.drivers.caiso_price.CaIsoPrice'
      })

    # add the feeds
    for (m, i, md) in self.MARKETS:
      for (f, u, fd) in self.FEEDS:
        path = '/%s/%s' % (m, f)
        self.add_timeseries(path, u, data_type='double',
                            description=md + ' ' + fd)

  def start(self):
    for (market, interval, description) in self.MARKETS:
      periodicSequentialCall(self.poll_stream, market, False).start(interval)

  def get_readings(self, market, start_date, stop_date): 
    readings = {'total_price': [], 'loss': [], 'energy': [], 'congestion': []}
    print "get_readings", market
    if market == 'DAM':
      q = 'PRC_LMP'
      m = 'DAM'
    elif market == 'HASP':
      q = 'PRC_HASP_LMP'
      m = 'HASP'
    elif market == 'RTM':
      q = 'PRC_INTVL_LMP'
      m = 'RTM'
    else:
      raise Exception("Invalid market: " + market)

    url = 'http://oasis.caiso.com/mrtu-oasis/SingleZip?'
    url += 'queryname=' + q
    url += '&startdate=' + dtutil.strftime_tz(start_date, '%Y%m%d', 'US/Pacific')
    url += '&enddate=' + dtutil.strftime_tz(stop_date, '%Y%m%d', 'US/Pacific')
    url += '&market_run_id=' + m
    url += '&node=' + self.location

    logging.info("Get url %s" % url)
    h = None
    for d in [5, 20, 60]:
      try:
        h = urllib2.urlopen(url, timeout=50)
        break
      except urllib2.URLError:
        logging.warn("urlopen failed.")
      time.sleep(d)
    if h == None:
      raise Exception("Failed to open url: %s" % url)

    z = zipfile.ZipFile(StringIO.StringIO(h.read()))
    xml = z.read(z.namelist()[0])
    b = BeautifulSoup.BeautifulSoup(xml)

    sec_per_int = int( b.find('m:sec_per_interval').contents[0] )

    rows = b.findAll('m:report_data')
    for d in rows:
      res = d.find('m:resource_name').contents[0]
      item = d.find('m:data_item').contents[0]
      day = d.find('m:opr_date').contents[0]
      inter = int( d.find('m:interval_num').contents[0] )
      val = float( d.find('m:value').contents[0] )

      secs = (inter - 1) * sec_per_int
      dt = dtutil.strptime_tz(day, '%Y-%m-%d', 'US/Pacific') + datetime.timedelta(seconds=secs)
      timestamp = dtutil.dt2ts(dt)

      key = None
      if item == 'LMP_PRC':
        key = 'total_price'
      elif item == 'LMP_LOSS_PRC':
        key = 'loss'
      elif item == 'LMP_ENE_PRC':
        key = 'energy'
      elif item == 'LMP_CONG_PRC':
        key = 'congestion'
      else:
        continue

      readings[key].append((timestamp, val))


    num_readings = len(readings[readings.keys()[0]])
    for k in readings.keys():
      if len(readings[k]) != num_readings:
        raise Exception('Missing readings')

      readings[k] = sorted(readings[k], key=lambda (t, v): t)

    return readings

  def poll_stream(self, market, load_old):
    def _push_data(readings, market): 
      # Zip together the values for all keys
      for vals in zip(*readings.values()):
        if vals[0][0] > self.last_reading[market]:
          # Add all smap points for this time
          for (k,v) in zip(readings.keys(), vals):
            logging.debug("add /%s/%s: %s" % (market, k, str(v)))
            self.add('/%s/%s' % (market, k), *v)
            self.last_reading[market] = vals[0][0]
            
    # Load old data
    if load_old == True:
      for day in range(1, 2):
        stop = dtutil.now() - datetime.timedelta(days=day)
        start = stop - datetime.timedelta(days=2)
        try:
          readings = self.get_readings(market, start, stop)
          _push_data(readings, market)
        except Exception, e:
          logging.exception('Error getting reading')

    # Continuously get new data
    try:
      stop = dtutil.now()
      start = stop - datetime.timedelta(days=1)

      readings = self.get_readings(market, start, stop)

      rt = readings['total_price'][-1][0]

      if rt > self.last_reading[market]:
        logging.info("NEW %s READING (%s) at time %s" %
                     (market, dtutil.strftime_tz(dtutil.ts2dt(rt), '%m/%d %H:%M', 'US/Pacific'),
                      dtutil.strftime_tz(dtutil.now(), '%m/%d %H:%M', 'US/Pacific')))
        _push_data(readings, market)
        # self.last_reading = rt

    except Exception, e:
      logging.exception('Error getting reading')




"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import os
import threading
import logging
import time
import socket

from twisted.internet.protocol import DatagramProtocol

# twisted doesn't support ipv6... this patches the reactor to add
# listenUDP6 and listenTCP6 methods.  It's not great, but it's a
# workaround that we can use and is easy to deploy (doesn't involve
# patching the twisted installation directory)
from tx.ipv6.internet import reactor

from smap.driver import SmapDriver
from smap.drivers.acmex2 import ACmeX2Report

class ACmeX2Driver(SmapDriver, DatagramProtocol):
    def datagramReceived(self, data, addr):
        rpt = ACmeX2Report.AcReport(data=data, data_length=len(data))
        moteid = addr[0].split('::')[1]
        if not moteid in self.ids:
            self.ids[moteid] = True
            self.add_timeseries('/' + moteid + '/true_power', 'mW', buffersz=2)
            self.add_timeseries('/' + moteid + '/apparent_power', 'mVA', buffersz=2)
            self.add_timeseries('/' + moteid + '/true_energy', 'mWh', buffersz=2)
            self.set_metadata('/' + moteid, {
                'Instrument/PartNumber' : moteid,
                'Instrument/SerialNumber' : ':'.join(['%02x' % x for x in rpt.get_eui64()]),
                'Instrument/SamplingPeriod' : str(rpt.get_period ()),
                })
        for idx in range(0,2):
            readingTime = rpt.get_globalTime() - (rpt.get_period() * (1 - idx))
            self._add('/' + moteid + '/true_energy', readingTime,
                      rpt.get_readings_cumulativeRealEnergy()[idx],
                      rpt.get_seq())
            self._add('/' + moteid + '/true_power', readingTime,
                      rpt.get_readings_averageRealPower()[idx],
                      rpt.get_seq())
            self._add('/' + moteid + '/apparent_power', readingTime,
                      rpt.get_readings_averageApparentPower()[idx],
                      rpt.get_seq())

    def setup(self, opts):
        self.port = int(opts.get('Port', 7002))
        self.ids = {}
        self.set_metadata('/', {
            'Extra/Driver' : 'smap.driver.acmex2.acmex2.ACmeX2Driver',
            'Instrument/Manufacturer': 'UC Berkeley',
            'Instrument/Model' : 'ACme X2',
            })

    def start(self):
        reactor.listenUDP6(self.port, self)



"""
Copyright (c) 2011, 2012, Regents of the University of California
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
@author Stephen Dawson-Haggerty <stevedh@eecs.berkeley.edu>
"""

import sys
import pycurl
import cStringIO as StringIO
import time
from threading import Thread
from Queue import Queue

from smap.util import json_decode

class ParserThread(Thread):
    """Parse the http results in parallel with getting them from the server
    """
    def __init__(self, parser):
        self.parser = parser
        self.q = Queue()
        self.ptime = 0
        self.rawlength = 0
        Thread.__init__(self)
    
    def run(self):
        self.result = []
        while True:
            next = self.q.get(True)
            if next == None: break
            (url, item) = next
            # print "parsing result for", url
            item.seek(0)
            tic = time.time()
            val = self.parser(item.read())
            self.ptime += (time.time() - tic)
            self.rawlength += item.tell()

            self.result.append((url, val))

    def add(self, url, pval):
        self.q.put_nowait((url, pval))

    def finish(self):
        self.q.put_nowait(None)
        self.join()
        return self.result

def mkrequest(c, spec):
    c.url = spec
    c.body = StringIO.StringIO()
    c.http_code = -1
    c.setopt(pycurl.URL, c.url)
    c.setopt(pycurl.WRITEFUNCTION, c.body.write)
    return c

def get(getspec, nconns=5, parser=json_decode, select_timeout=1.0, verbose=True):
    """get a list of urls, using a connection pool of up to nconn connections.
    apply "parser" to each of the results.

    Based on retriever-multi.py.
    """
    tic = time.time()
    parser_thread = ParserThread(parser)
    parser_thread.start()

    rv = []
    m = pycurl.CurlMulti()
    m.handles = []
    for spec in xrange(nconns):
        c = pycurl.Curl()
        c.fp = None
        c.setopt(pycurl.FOLLOWLOCATION, 1)
        c.setopt(pycurl.MAXREDIRS, 5)
        c.setopt(pycurl.CONNECTTIMEOUT, 30)
        c.setopt(pycurl.TIMEOUT, 300)
        c.setopt(pycurl.NOSIGNAL, 1)
        m.handles.append(c)

    freelist = m.handles[:]
    num_processed, num_urls = 0, len(getspec)
    while num_processed < num_urls:
        while getspec and freelist:
            spec = getspec.pop(0)
            c = freelist.pop()
            mkrequest(c, spec)
            m.add_handle(c)

        while 1:
            ret, num_handles = m.perform()
            if ret != pycurl.E_CALL_MULTI_PERFORM: 
                break

        while 1:
            num_q, ok_list, err_list = m.info_read()
            for c in ok_list:
                # print "Success:", c.url, c.getinfo(pycurl.EFFECTIVE_URL)
                # print "Success:", c.url, c.getinfo(pycurl.EFFECTIVE_URL)
                parser_thread.add(c.url, c.body)
                c.fp = None
                c.body = None
                m.remove_handle(c)
                freelist.append(c) 

            for c, errno, errmsg in err_list:
                m.remove_handle(c)
                print "Failed: ", c.url, errno, errmsg
                freelist.append(c)

            num_processed += len(ok_list) + len(err_list)
            if verbose:
                print >>sys.stderr, str(num_processed) + '/' + str(num_urls) +  '\r'
            if num_q == 0:
                break

        m.select(select_timeout)
    print

    for c in m.handles:
        c.close()
    m.close()
    dlend = time.time()

    rv = parser_thread.finish()
    toc = time.time()
    print """downloaded %ib from %i urls in %.03fs (%.03fMB/s download: %.03fs, parse: %03fs)""" % \
        (parser_thread.rawlength, 
         
         len(rv), toc - tic, 
         float(parser_thread.rawlength) / ((dlend - tic) * 1e6),
         dlend - tic, parser_thread.ptime)
    return rv



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
# See LICENSE for details.
#
# Exceedingly minor changes by B. A. Shelton, zancarius@gmail.com
# !!! Do not blame Twisted Matrix Laboratories if using this monkey patch breaks
#     your application. !!!
#
# This file is based entirely on the twisted.internet.tcp module as it ships
# with Twisted v10.1. Only very minor changes have been made, specifically
# all references to TCP and UDP have been moved to TCP6/UDP6 and all IPv4
# references in the comments and the code have been moved to IPv6. Socket
# address families have been bumped to socket.AF_INET6.
#
# Class changes in this file are implemented by subclassing the appropriate
# Twisted classes. Note that the Port implementation in this module (Port6)
# forces the socket to listen only on IPv6 connections; this should allow you
# to use IPv4 and IPv6 simultaneously.

import socket
import sys

from twisted.python.runtime import platformType

from twisted.internet import abstract
from twisted.internet import base
from twisted.internet import tcp
from tx.ipv6.internet import address6 as address

# Twisted imports for overrides.
import twisted.internet.tcp

class BaseClient(twisted.internet.tcp.BaseClient):
    """A base class for client TCP (and similiar) sockets.
    """
    addressFamily = socket.AF_INET6
    socketType = socket.SOCK_STREAM

    def resolveAddress(self):
        if abstract.isIPAddress(self.addr[0]):
            self._setRealAddress(self.addr[0])
        else:
            d = self.reactor.resolve(self.addr[0])
            d.addCallbacks(self._setRealAddress, self.failIfNotConnected)

    def _setRealAddress(self, address):
        self.realAddress = (address, self.addr[1])
        self.doConnect()

class Server(twisted.internet.tcp.Server):
    """
    Serverside socket-stream connection class.

    This is a serverside network connection transport; a socket which came from
    an accept() on a server.
    """

    def getHost(self):
        """Returns an IPv6Address.

        This indicates the server's address.
        """
        return address.IPv6Address('TCP6', *(self.socket.getsockname() + ('INET6',)))

    def getPeer(self):
        """Returns an IPv6Address.

        This indicates the client's address.
        """
        return address.IPv6Address('TCP6', *(self.client + ('INET6',)))

class Port6(tcp.Port):

    addressFamily = socket.AF_INET6
    socketType = socket.SOCK_STREAM

    # Do not remove this. You will break something.
    transport = Server

    def createInternetSocket(self):
        s = base.BasePort.createInternetSocket(self)
        if platformType == "posix" and sys.platform != "cygwin":
            # Required: Forces listenTCP6 to listen exclusively on IPv6 addresses.
            # See: http://www.velocityreviews.com/forums/t328345-ipv6-question.html
            s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s

    def _buildAddr(self, (host, port, flowinfo, scopeid)):
        return address._ServerFactoryIPv6Address('TCP6', host, port, flowinfo, scopeid)

    def getHost(self):
        """Returns an IPv6Address.

        This indicates the server's address.
        """
        # IPv6 returns: (address, port, flowinfo, scopeid) for socket.getsockname()
        return address.IPv6Address('TCP6', *(self.socket.getsockname() + ('INET6',)))
